"""Reader for Biognosys Spectronaut TSV spectral libraries"""

import json
import os

from typing import List, Tuple, Dict, Iterator, Any, Deque, Union

from pyteomics import proforma

from mzspeclib import annotation
from mzspeclib.analyte import Analyte
from mzspeclib.backends.base import LIBRARY_NAME, _CSVSpectralLibraryBackendBase, FORMAT_VERSION, DEFAULT_VERSION
from mzspeclib.backends.utils import open_stream, urlify, try_cast
from mzspeclib.spectrum import Spectrum, SPECTRUM_NAME
from mzspeclib import const as c
from mzspeclib.const import (
    CHARGE_STATE,
    SOURCE_FILE,
    STRIPPED_PEPTIDE_SEQ as STRIPPED_PEPTIDE_TERM,
    PROFORMA_SEQ as PROFORMA_PEPTIDE_TERM,
    CUSTOM_ATTRIBUTE_NAME,
    CUSTOM_ATTRIBUTE_VALUE,
)


SELECTED_ION_MZ = "MS:1003208|experimental precursor monoisotopic m/z"

ID_SEP = "/"
_ID_SEP = ID_SEP.encode("ascii")

NO_LOSS = 'noloss'


def _rewrite_modified_peptide_as_proforma(sequence: str) -> str:
    if sequence.startswith("_"):
        sequence = sequence.strip("_")
    buffer: Deque[str] = Deque()
    last_paren = None
    for i, c in enumerate(sequence):
        if c == ']':
            # Erase any text in parentheses as these indicate the modification
            # rule and not the modificatin name. We could look at the modification
            # rule to infer N-term and C-term rules, but we don't have enough examples
            if last_paren is not None:
                k = i - last_paren
                for _ in range(k + 1):
                    buffer.pop()
                last_paren = None
                buffer.append(c)
        elif c == '(':
            last_paren = i
            buffer.append(c)
        else:
            buffer.append(c)
    pf_seq = ''.join(buffer)
    # A peptide with an N-terminal modification will start with a square brace
    # but needs to have a "-" added to be well-formed ProForma
    if pf_seq.startswith("["):
        i = pf_seq.find(']') + 1
        if i == 0:
            raise ValueError(f"Malformed peptide sequence {sequence}")
        pf_seq = f"{pf_seq[:i]}-{pf_seq[i:]}"
    return pf_seq


def _parse_value(value: str) -> Union[float, int, str, bool]:
    try:
        return json.loads(value)
    except json.JSONDecodeError:
        lower = value.lower()
        if lower == "true":
            return True
        elif lower == "false":
            return False
        return value


class SpectronautTSVSpectralLibrary(_CSVSpectralLibraryBackendBase):
    """Read Spectronaut TSV spectral libraries."""

    format_name = "spectronaut.tsv"

    _custom_spectrum_keys = [
        "ExcludeFromAssay",
        "BGSInferenceId",
        "AllowForNormalization",
        "Workflow",
    ]

    _custom_analyte_keys = [
        "IsProteotypic",
        "FASTAName",
        "Database",
        "ProteinGroups",
    ]

    _key_columns = ['ModifiedPeptide', 'PrecursorCharge']

    _required_columns = ['PrecursorMz', 'PrecursorCharge',
                         'ModifiedPeptide', 'StrippedPeptide',
                         'FragmentMz', 'RelativeIntensity',
                         'FragmentType', 'FragmentNumber',
                         'FragmentCharge', 'FragmentLossType']

    def __init__(self, filename: str, index_type=None, **kwargs):
        super().__init__(filename, index_type=index_type, delimiter='\t', **kwargs)

    def _spectrum_origin_type(self):
        key = c.SPECTRUM_ORIGIN_TYPE
        value = c.SELECTED_FRAGMENT_THEORETICAL_OBSERVED_INTENSITY_SPECTRUM
        return key, value

    def _spectrum_aggregation_type(self):
        key = c.SPECTRUM_AGGREGATION_TYPE
        value = c.CONSENSUS_SPECTRUM
        return key, value

    def read_header(self) -> bool:
        result = super().read_header()
        self.add_attribute(FORMAT_VERSION, DEFAULT_VERSION)
        if hasattr(self.filename, 'name'):
            name = self.filename.name.replace(".gz", '').rsplit('.', 1)[0].split(os.sep)[-1]
        else:
            name = self.filename.replace(".gz", '').rsplit(".", 1)[0].split(os.sep)[-1]
        self.add_attribute(LIBRARY_NAME, name)
        self.add_attribute(c.LIBRARY_CREATION_SW, "MS:1001327|Spectronaut")
        return result

    def _batch_rows(self, iterator: Iterator[Dict[str, Any]]) -> Iterator[List[Dict[str, Any]]]:
        group_key = None
        group = []
        i = 0
        for row in iterator:
            key = [row[k_i] for k_i in self._key_columns]
            if group_key is None:
                group_key = key
                group.append(row)
            elif group_key == key:
                group.append(row)
            else:
                i += 1
                yield group
                group = [row]
                group_key = key
        if group:
            yield group

    def create_index(self):
        key = None
        delimiter = self._delimiter.encode("ascii")
        with open_stream(self.filename, 'rb') as stream:
            header = stream.readline()
            header_cols = header.split(delimiter)
            column_keys = (
                header_cols.index(b'ModifiedPeptide'),
                header_cols.index(b'PrecursorCharge'),
            )
            offset = stream.tell()

            line = stream.readline()
            tokens = line.split(delimiter)
            key = (tokens[column_keys[0]].strip(b"_"), tokens[column_keys[1]])
            n = 0
            self.index.add(
                number=n,
                offset=offset,
                name=_ID_SEP.join(key).decode("utf8"),
                analyte=None
            )
            n += 1

            # To hold previous values
            last_offset = None
            while line:
                tokens = line.split(delimiter)
                next_key = (tokens[column_keys[0]].strip(b"_"), tokens[column_keys[1]])
                if next_key != key:
                    key = next_key
                    self.index.add(
                        number=n,
                        offset=offset,
                        name=_ID_SEP.join(key).decode("utf8"),
                        analyte=None
                    )
                    n += 1
                    last_offset = stream.tell()
                else:
                    offset = stream.tell()
                line = stream.readline()
        n += 1
        self.index.commit()
        return n

    def _generate_peaks(self, batch: List[Dict[str, Any]]) -> List[Tuple[float, float, List[annotation.IonAnnotationBase], List]]:
        peaks = []
        for row in batch:
            mz = float(row['FragmentMz'])
            intensity = float(row['RelativeIntensity'])

            series = row['FragmentType']
            ordinal = int(row['FragmentNumber'])
            charge = int(row['FragmentCharge'])

            loss_type = row['FragmentLossType']
            if loss_type != NO_LOSS:
                loss_type = annotation.NeutralName.parse('-' + loss_type)
            else:
                loss_type = None

            annot = annotation.PeptideFragmentIonAnnotation(
                series, ordinal, neutral_losses=loss_type, charge=charge,
                mass_error=annotation.MassError(0, 'Da')
            )

            peak = [
                mz, intensity, [annot], []
            ]
            peaks.append(peak)
        return peaks

    def _build_analyte(self, description: Dict[str, Any], analyte: Analyte) -> Analyte:
        pf_seq = _rewrite_modified_peptide_as_proforma(description['ModifiedPeptide'])
        peptide = proforma.ProForma.parse(pf_seq)
        peptide.charge_state = int(description['PrecursorCharge'])
        analyte.add_attribute(STRIPPED_PEPTIDE_TERM, description['StrippedPeptide'])
        analyte.add_attribute(c.PROFORMA_ION, str(peptide))
        analyte.add_attribute(c.THEORETICAL_MASS, peptide.mass)


        protein_group_id = analyte.get_next_group_identifier()


        if 'UniProtIds' in description:
            analyte.add_attribute(
                c.PROTEIN_ACCESSION,
                description['UniProtIds'],
                group_identifier=protein_group_id
            )
        if 'Protein Name' in description:
            analyte.add_attribute(
                c.PROTEIN_NAME,
                description["Protein Name"],
                group_identifier=protein_group_id
            )
        if 'ProteinDescription' in description:
            analyte.add_attribute(
                c.PROTEIN_DESCRIPTION,
                description['ProteinDescription'],
                group_identifier=protein_group_id
            )

        if "OrganismId" in description and description["OrganismId"] is not None:
            analyte.add_attribute_group([
                [c.TAXONOMY_NCBI_TAX_ID, try_cast(description['OrganismId'])],
                [c.TAXONOMY_SCIENTIFIC_NAME, description['Organisms']],
            ])

        for key in self._custom_analyte_keys:
            if key in description:
                analyte.add_attribute_group([
                    [CUSTOM_ATTRIBUTE_NAME, key],
                    [CUSTOM_ATTRIBUTE_VALUE, _parse_value(description[key])]
                ])

        return analyte

    def _parse_from_buffer(self, buffer: List[Dict[str, Any]], spectrum_index: int = None) -> Spectrum:
        spec = self._new_spectrum()
        descr = buffer[0]

        key = (descr['ModifiedPeptide'].strip("_"), descr['PrecursorCharge'])

        spec.add_attribute(SPECTRUM_NAME, ID_SEP.join(key))
        spec.add_attribute(SELECTED_ION_MZ, float(descr['PrecursorMz']))
        spec.add_attribute(CHARGE_STATE, int(descr['PrecursorCharge']))
        spec.add_attribute(SOURCE_FILE, urlify(descr['ReferenceRun']))
        spec.add_attribute(*self._spectrum_origin_type())
        spec.add_attribute(*self._spectrum_aggregation_type())

        spec.add_attribute_group([
            [CUSTOM_ATTRIBUTE_NAME, "LabeledPeptide"],
            [CUSTOM_ATTRIBUTE_VALUE, descr['LabeledPeptide']]
        ])

        if 'IonMobility' in descr:
            spec.add_attribute("MS:1002476|ion mobility drift time", float(descr['IonMobility']))
        if 'CV' in descr:
            spec.add_attribute("MS:1001581|FAIMS compensation voltage", float(descr['CV']))
        if 'iRT' in descr:
            group_id = spec.get_next_group_identifier()
            spec.add_attribute("MS:1000896|normalized retention time", float(descr['iRT']), group_identifier=group_id)
            spec.add_attribute(
                c.UNIT,
                c.MINUTE,
                group_identifier=group_id,
            )

        analyte = self._new_analyte('1')
        self._build_analyte(descr, analyte)
        spec.add_analyte(analyte)

        for key in self._custom_spectrum_keys:
            if key in descr:
                spec.add_attribute_group([
                    [CUSTOM_ATTRIBUTE_NAME, key],
                    [CUSTOM_ATTRIBUTE_VALUE, _parse_value(descr[key])]
                ])

        spec.peak_list = self._generate_peaks(buffer)
        spec.add_attribute(c.NUM_PEAKS, len(spec.peak_list))

        if spectrum_index is not None:
            spec.index = spectrum_index
        else:
            spec.index = -1
        spec.key = spec.index + 1
        return spec



import unittest

from mzspeclib.backends import (MSPSpectralLibrary)
from mzspeclib.backends import msp

from .common import datafile


class TestMSPAttributeConversion(unittest.TestCase):
    def test_hierarchy(self):
        flat_maps = (msp.analyte_terms, msp.other_terms, msp.interpretation_terms, msp.interpretation_member_terms)
        handler_maps = (msp.msp_analyte_attribute_handler, msp.msp_spectrum_attribute_handler)
        for handler_map in handler_maps:
            for key in handler_map.mapping:
                for flat_map in flat_maps:
                    assert key not in flat_map



"""HUPO-PSI Spectral library format."""

from mzspeclib.spectrum import Spectrum
from mzspeclib.analyte import Analyte, Interpretation, InterpretationMember, InterpretationCollection
from mzspeclib.cluster import SpectrumCluster

from mzspeclib.index import MemoryIndex, SQLIndex

from mzspeclib.spectrum_library import SpectrumLibrary
from mzspeclib.spectrum_library_index import SpectrumLibraryIndex
from mzspeclib.spectrum_library_collection import SpectrumLibraryCollection
from mzspeclib.universal_spectrum_identifier import UniversalSpectrumIdentifier



#!/usr/bin/env python3
import os
import pathlib

from typing import Any, Iterator, Optional, Set, Type, List, Union
from mzspeclib.attributes import Attribute, AttributeManagedProperty, AttributeManager
from mzspeclib.backends.base import LIBRARY_DESCRIPTION, LIBRARY_NAME, LIBRARY_URI, LIBRARY_VERSION
from mzspeclib.cluster import SpectrumCluster

from mzspeclib.spectrum_library_index import SpectrumLibraryIndex
from mzspeclib.spectrum import Spectrum
from mzspeclib.index import IndexBase
from mzspeclib.backends import guess_implementation, SpectralLibraryBackendBase, SpectralLibraryWriterBase, FormatInferenceFailure


class SpectrumLibrary:
    """
    Read, write, and search through a spectrum library.

    This type will attempt to infer the correct spectrum library reader from the input
    file, but may need to be explicitly prompted if there are ambiguities.

    Attributes
    ----------
    identifier: str
        A unique identifier string assigned to this library by a spectral library host or
        provider.
    filename: str or Path
        A location on the local file system where the spectral library is stored
    format : str
        The name of the format for the current encoding of the library.
    backend: :class:`~.SpectralLibraryBackendBase`
        The implementation used to parse the file

    Parameters
    ----------
    identifier : str, optional
        A universal identifier for a hosted spectral library to fetch.
    filename : str, os.PathLike, or io.IOBase, optional
        A path-like or file-like object that holds a spectral library to read.
    format : string
        Name of the format for the current encoding of the library.
    index_type : Type[:class:`~.mzlib.index.base.IndexBase`]
        The type of index to preferentially construct.
    create_index : bool
        Whether to construct an index over the library if one does not exist
        already. This limits the library to sequential iteration but does not
        incur an expensive end-to-end parsing step upon opening.
    """

    backend: SpectralLibraryBackendBase
    filename: str
    identifier: str
    format: str
    index_type: Type[IndexBase]

    name = AttributeManagedProperty[str](LIBRARY_NAME)
    description = AttributeManagedProperty[str](LIBRARY_DESCRIPTION)
    uri = AttributeManagedProperty[str](LIBRARY_URI)
    library_version = AttributeManagedProperty[str](LIBRARY_VERSION)

    _identifier = None
    _filename = None
    _create_index: bool

    def __repr__(self):
        components = []

        if self.identifier:
            components.append(f"identifier={self.identifier!r}")
        if self.filename:
            components.append(f"filename={self.filename!r}")
        if self.backend is not None:
            components.append(f"backend={self.backend.__class__.__name__}")

        return f"{self.__class__.__name__}({', '.join(components)})"

    def __init__(
        self,
        identifier: Optional[str] = None,
        filename: Optional[str] = None,
        format: Optional[Type[SpectralLibraryBackendBase]]=None,
        index_type: Optional[Type[IndexBase]]=None,
        create_index: bool=True,
    ):
        self.backend = None
        self._create_index = create_index
        self._format = format
        self.identifier = identifier
        self.index_type = index_type
        self.filename = filename

    def _init_from_identifier(self):
        if os.path.exists(self.identifier) and self._filename is None and not self._backend_initialized():
            # This in turn initializes the backend
            self.filename = self.identifier

    def _init_from_filename(self, index_type: Optional[Type[IndexBase]]=None):
        if index_type is None:
            index_type = self.index_type
        if self.format is None:
            self.backend = guess_implementation(self.filename, index_type)
            self._format = self.backend.format_name
        else:
            if callable(self.format):
                backend_type = self.format
            else:
                backend_type = SpectralLibraryBackendBase.type_for_format(self.format)
            if backend_type is None:
                raise ValueError(
                    f"Could not find an implementation for {self.format}")
            self.backend = backend_type(
                self.filename, index_type=index_type, create_index=self._create_index)
            self._format = self.backend.format_name
        self._identifier = self.backend.identifier

    def _backend_initialized(self):
        return self.backend is not None

    def _requires_backend(self):
        if not self._backend_initialized():
            raise ValueError(
                "Cannot read library data, library parser not yet initialized")

    @classmethod
    def from_backend(cls, backend: SpectralLibraryBackendBase) -> "SpectrumLibrary":
        """
        Wrap a format-specific backend in a :class:`SpectrumLibrary`.

        This is useful because the respective backends do not support all
        operations.
        """
        self = cls()
        self.backend = backend
        self._format = backend.__class__
        self.filename = self.backend.filename
        self.index_type = self.backend.index.__class__
        return self

    @property
    def spectrum_attribute_sets(self):
        """The spectrum attribute sets of the spectral library"""
        self._requires_backend()
        return self.backend.spectrum_attribute_sets

    @property
    def analyte_attribute_sets(self):
        """The analyte attribute sets of the spectral library"""
        self._requires_backend()
        return self.backend.analyte_attribute_sets

    @property
    def interpretation_attribute_sets(self):
        """The interpretation attribute sets of the spectral library"""
        self._requires_backend()
        return self.backend.interpretation_attribute_sets

    @property
    def cluster_attribute_sets(self):
        """The spectrum cluster attribute sets of the spectral library"""
        self._requires_backend()
        return self.backend.cluster_attribute_sets

    #### Define getter/setter for attribute identifier
    @property
    def identifier(self) -> Optional[str]:
        """
        The spectrum library's identifier, such an accession number provided by a host or repository.

        This attribute may not exist for libraries loaded from files on disk. It relies on the presence
        of the ``MS:1003187|library identifier`` attribute in the library header.
        """
        if self._identifier is None:
            if self._backend_initialized():
                return self.backend.identifier
        return self._identifier

    @identifier.setter
    def identifier(self, identifier: Optional[str]):
        self._identifier = identifier
        if self.backend is not None:
            self.backend.identifier = identifier

    #### Define getter/setter for attribute filename
    @property
    def filename(self) -> Optional[str]:
        """
        The path to the library file on disk.

        This may also be a file-like object.
        """
        return self._filename

    @filename.setter
    def filename(self, filename):
        self._filename = filename
        if filename is not None:
            self._init_from_filename()

    #### Define getter/setter for attribute format
    @property
    def format(self):
        return self._format

    @property
    def index(self) -> IndexBase:
        if self._backend_initialized():
            return self.backend.index
        return None

    @property
    def attributes(self) -> AttributeManager:
        """The library level attributes"""
        if self._backend_initialized():
            return self.backend.attributes
        return None

    def read_header(self) -> bool:
        """
        Read just the header of the whole library

        Returns
        -------
        bool:
            Whether the operation was successful
        """
        self._requires_backend()
        return self.backend.read_header()

    def read(self):
        """
        Create a sequential iterator over the spectrum library entries.

        Yields
        ------
        Spectrum or SpectrumCluster
        """
        self._requires_backend()
        return self.backend.read()

    def write(self, destination, format: str=None, **kwargs):
        """
        Write the library to disk.

        Parameters
        ----------
        destination : str, os.PathLike, or io.IOBase
            The path or stream to write the library to.
        format : str, Type, or Callable
            The name of the format or a callable object that returns
            a :class:`~.SpectrumLibraryWriterBase`.
        **kwargs
            Passed to implementation.
        """
        filename = destination
        if not isinstance(filename, (str, pathlib.Path, os.PathLike)):
            filename = getattr(destination, "name", None)

        if format is None and filename is not None:
            basename = os.path.basename(filename)
            tokens = basename.rsplit(".", 2)
            if len(tokens) == 3:
                writer_type = SpectralLibraryWriterBase.type_for_format('.'.join(tokens[1:]))
            else:
                raise ValueError(f"Could not guess format from file name {filename}")
        else:
            writer_type = SpectralLibraryWriterBase.type_for_format(format)
        if writer_type is None:
            raise ValueError(
                f"Could not find a format writer from file name {filename} or format {format}")
        writer = writer_type(destination, **kwargs)
        if self._backend_initialized():
            with writer:
                writer.write_library(self.backend)
        else:
            print("Library not initialized")
            writer.close()

    def get_spectrum(self, spectrum_number: int=None, spectrum_name: str=None) -> Spectrum:
        """
        Retrieve a single spectrum from the library.

        Parameters
        ----------
        spectrum_number : int, optional
            The index of the specturm in the library
        spectrum_name : str, optional
            The name of the spectrum in the library

        Returns
        -------
        :class:`~.Spectrum`
        """
        self._requires_backend()
        return self.backend.get_spectrum(spectrum_number, spectrum_name)

    def get_cluster(self, cluster_number: int) -> SpectrumCluster:
        """
        Retrieve a single spectrum cluster from the library.

        Parameters
        ----------
        cluster_number : int, optional
            The index of the cluster in the library

        Returns
        -------
        :class:`~.SpectrumCluster`
        """
        self._requires_backend()
        return self.backend.get_cluster(cluster_number)

    def find_spectra(self, specification, **query_keys) -> List[Spectrum]:
        """
        Return a list of spectra given query constraints

        Returns
        -------
        List[Spectrum]
        """
        self._requires_backend()
        return self.backend.find_spectra(specification, **query_keys)

    def __getitem__(self, i: int) -> Spectrum:
        self._requires_backend()
        return self.backend[i]

    def __len__(self):
        if self._backend_initialized():
            return len(self.backend)
        return 0

    def __iter__(self) -> Iterator[Spectrum]:
        if self._backend_initialized():
            return iter(self.backend)
        return iter([])

    def add_attribute(self, key, value, group_identifier=None):
        """
        Add an attribute to the library level attributes store.

        Parameters
        ----------
        key : str
            The name of the attribute to add
        value : object
            The value of the attribute to add
        group_identifier : str, optional
            The attribute group identifier to use, if any. If not provided,
            no group is assumed.
        """
        self._requires_backend()
        return self.backend.add_attribute(key, value, group_identifier=group_identifier)

    def get_attribute(
        self, key: str, group_identifier: Optional[str] = None, raw: bool = False
    ) -> Union[Any, List[Any], Attribute, List[Attribute]]:
        """
        Get the value or values associated with a given
        attribute key.

        Parameters
        ----------
        key : str
            The name of the attribute to retrieve
        group_identifier : str, optional
            The specific group identifier to return from.
        raw : bool
            Whether to return the :class:`Attribute` object or unwrap the value

        Returns
        -------
        attribute_value: object or list[object]
            Returns single or multiple values for the requested attribute.
        """
        self._requires_backend()
        return self.backend.get_attribute(key, group_identifier=group_identifier, raw=raw)

    def remove_attribute(self, key, group_identifier=None):
        """
        Remove the value or values associated with a given
        attribute key from the library level attribute store.

        This rebuilds the entire store, which may be expensive.

        Parameters
        ----------
        key : str
            The name of the attribute to retrieve
        group_identifier : str, optional
            The specific group identifier to return from.

        """
        self._requires_backend()
        return self.backend.remove_attribute(key, group_identifier=group_identifier)

    def has_attribute(self, key):
        """
        Test for the presence of a given attribute in the library
        level store.

        Parameters
        ----------
        key : str
            The attribute to test for

        Returns
        -------
        bool
        """
        self._requires_backend()
        return self.backend.has_attribute(key)

    def summarize_parsing_errors(self):
        """Retrieve a free-form description of parsing errors"""
        return self.backend.summarize_parsing_errors()

    @staticmethod
    def supported_file_extensions() -> Set[str]:
        """
        Get the set of file extensions that are currently recognized as spectral
        library formats
        """
        return set(SpectralLibraryBackendBase._file_extension_to_implementation)


#### Example using this class
def example():

    #### Create a new RTXFeedback object
    spectrum_library = SpectrumLibrary()
    spectrum_library.filename = "../refData/sigmaups1_consensus_final_true_lib.msp"
    spectrum_library.read_header()

    spectrum_buffer = spectrum_library.get_spectrum(spectrum_index_number=2000)
    spectrum = Spectrum()
    spectrum.parse(spectrum_buffer)
    buffer = spectrum.write(format="text")
    print(buffer)
    print()

    return()


#### If this class is run from the command line, perform a short little test to see if it is working correctly
def main():

    #### Run an example
    example()
    return()


if __name__ == "__main__": main()




import logging
import sys

from typing import List, DefaultDict

from mzspeclib import SpectrumLibrary
from mzspeclib.validate import *

logging.basicConfig(level=logging.DEBUG, stream=sys.stderr)
logger = logging.getLogger()

library = SpectrumLibrary(filename="examples/fetal_brain_tiny.mzlib.txt")

chain = load_default_validator()
chain.validate_library(library)

by_level: DefaultDict[RequirementLevel, List[ValidationError]] = DefaultDict(list)
for message in chain.error_log:
    by_level[message.requirement_level].append(message)

for level, bucket in sorted(by_level.items()):
    log_level = logging.WARN
    if level == RequirementLevel.may:
        log_level = logging.DEBUG
    logger.log(log_level, f"Found {len(bucket)} violations for {level.name.upper()} rules")
    for err in bucket:
        logger.log(log_level, f"... {err.message}")


import json
from pathlib import Path

import pandas as pd
from jsonschema import validate


def excel_terms_to_dict(excel_file, sheet_name):
    terms_sheet = pd.read_excel(excel_file, sheet_name=sheet_name)
    # Remove trailing spaces from all dataframe values
    terms_sheet = terms_sheet.apply(
        lambda x: x.str.strip() if x.dtype == "object" else x
    )
    terms = terms_sheet.to_dict(orient="records")
    return terms


def parse_terms(terms):
    """Translate terms from excel to metadata rules."""
    for term in terms:
        attribute = {
            "accession": term["accession"],
            "name": term["name"],
            "repeatable": term["repeatable"],
        }

        if isinstance(term["notes"], str):
            attribute["notes"] = term["notes"]

        if term["value"].startswith("Children of"):
            attribute["value"] = {
                "name": "value_is_child_of",
                "accession": term["value"].split("Children of ")[1].strip(),
            }
        elif term["value"].startswith("xsd:"):
            attribute["value"] = {"name": "value_of_type", "value": term["value"]}
        yield {
            "id": "has_" + term["name"].lower().replace(" ", "_"),
            "path": term["path"],
            "attr": [attribute],
            "requirement_level": term["level"],
        }


def main(schema_file, excel_file, sheet_name, json_file):
    # Read Excel
    terms = excel_terms_to_dict(excel_file, sheet_name)

    # Parse terms to validation rules
    rules = {
        "$schema": "./validator-rules-schema.json",
        "name": "all_metadata",
        "rules": [t for t in parse_terms(terms)],
    }

    # Validate
    validate(rules, json.load(open(schema_file)))

    # Write JSON
    with open(json_file, "w") as f:
        json.dump(rules, f, indent=2)


if __name__ == "__main__":
    current_path = Path(__file__).parent
    schema_file = current_path / "../validator-rules-schema.json"
    excel_file = current_path / "terms.xlsx"
    sheet_name = "general"
    json_file = current_path / "../all.json"

    main(schema_file, excel_file, sheet_name, json_file)



"""Tools for interacting with controlled vocabularies and biomedical ontologies."""

import logging
from typing import Dict

from psims.controlled_vocabulary import Entity, ControlledVocabulary
from psims.controlled_vocabulary.controlled_vocabulary import load_uo, load_unimod, load_psims

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class _VocabularyResolverMixin(object):
    """A base class for looking up terms in a set of controlled vocabularies."""

    default_cv_loader_map = {
        "MS": load_psims,
        "UO": load_uo,
        "UNIMOD": load_unimod,
    }

    controlled_vocabularies: Dict[str, ControlledVocabulary]

    def __init__(self, *args, **kwargs):
        self.controlled_vocabularies = dict()
        super().__init__(*args, **kwargs)

    def load_cv(self, name: str) -> ControlledVocabulary:
        """
        Retrieve a controlled vocabulary by name.

        This may query the internet or on-disk cache when called for the
        first time for each name, and the result is cached in memory.
        There may be inconsistencies from one source to the next.

        Parameters
        ----------
        name : str
            The name to load

        Returns
        -------
        :class:`psims.controlled_vocabular.ControlledVocabulary`
        """
        if name in self.controlled_vocabularies:
            return self.controlled_vocabularies[name]
        self.controlled_vocabularies[name] = self.default_cv_loader_map[name]()
        return self.controlled_vocabularies[name]

    def find_term_for(self, curie: str) -> Entity:
        try:
            name, _id = curie.split(":", 1)
        except ValueError as err:
            raise ValueError(f"Could not parse {curie} into source and accession. Is it a CURIE?") from None
        cv = self.load_cv(name)
        term = cv[curie]
        return term

    def find_term_by_name(self, name: str) -> Entity:
        cv_keys = ['MS', 'UNIMOD']
        for cv_key in cv_keys:
            cv = self.load_cv(cv_key)
            try:
                term = cv[name]
                return term
            except KeyError:
                continue
        raise KeyError(name)


class ControlledVocabularyResolver(_VocabularyResolverMixin):

    def name_to_curie(self, name: str) -> str:
        term = self.find_term_by_name(name)
        return term.id

    def attribute_syntax(self, name: str) -> str:
        """Given an identifier, format it for textual representation"""
        if self.is_curie(name):
            if "|" in name:
                return name
            term = self.find_term_for(name)
        else:
            term = self.find_term_by_name(name)
        return f"{term.id}|{term.name}"

    def is_curie(self, name: str) -> bool:
        """Test if a string is a CURIE"""
        if ':' not in name:
            return False
        prefix, _rest = name.split(":")
        if prefix in self.default_cv_loader_map:
            return True
        return False



"""
.. Controlled Vocabulary Attributes
.. ================================

Like many PSI file formats, :title-reference:`MzSpecLib` uses controlled vocabulary
terms to represent properties of entities. Many places in :mod:`mzspeclib`
use :class:`~.Attribute` to evaluate spectra, but they are also used indirectly
when interacting with a :class:`~.AttributeManager` as a mapping for key-value pair
retrieval.
"""
import textwrap

from typing import (
    Any, DefaultDict, Iterable,
    Iterator, Optional, Set, Tuple,
    Union, List, Dict,
    Generic, TypeVar, Type
)

from dataclasses import dataclass

from . import const

T = TypeVar('T')


class Attribute(object):
    __slots__ = ("key", "value", "group_id", "owner_id")

    key: str
    value: Union[str, int, float, 'Attribute', List]
    group_id: Optional[str]
    owner_id: Optional[Any]

    def __init__(self, key, value, group_id=None, owner_id=None):
        self.key = key
        self.value = value
        self.group_id = group_id
        self.owner_id = owner_id

    def copy(self):
        """Create a shallow copy of this :class:`Attribute`"""
        return self.__class__(self.key, self.value, self.group_id, self.owner_id)

    def __getitem__(self, i):
        if i == 0:
            return self.key
        elif i == 1:
            return self.value
        elif i == 2:
            return self.group_id
        elif i == 3:
            return self.owner_id
        else:
            raise IndexError(i)

    def __iter__(self):
        yield self.key
        yield self.value
        yield self.group_id
        yield self.owner_id

    def __len__(self):
        if self.group_id is None:
            return 2
        return 3

    def __eq__(self, other):
        if other is None:
            return False
        return self.key == other.key and self.value == other.value and self.group_id == other.group_id

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        return f"{self.__class__.__name__}({self.key}, {self.value}, {self.group_id})"

    def __str__(self):
        base = f"{self.key}={self.value}"
        if self.group_id:
            base = f"[{self.group_id}]" + base
        return base

    def __hash__(self):
        key = (self.key, )
        if self.group_id is not None:
            key = (self.key, self.group_id)
        return hash(key)


class AttributeManager(object):
    """
    A key-value pair store with optional grouping for storing controlled
    vocabulary-backed attributes.

    The various components of this object shouldn't be modified directly,
    and instead rely on the interface this type provides to access its data.

    Parameters
    ----------
    attributes : Iterable[list], optional
        Attribute name-value pairs with an optional grouping value. If omitted,
        the attribute store will be empty.

    """

    # Attributes
    #     ----------
    #     attributes: List[List]
    #         The actual attribute name-value pairs with an optional grouping value
    #     attribute_dict: Dict
    #         A mapping from attribute name to indices into :attr:`attributes` and
    #         the group assignments.
    #     group_dict: Dict
    #         A mapping from group identifier to indices into :attr:`attributes`
    #     group_counter: int
    #         The number of attribute groups assigned.

    attributes: List[Attribute]
    attribute_dict: Dict
    group_dict: Dict
    group_counter: int


    __slots__ = ('attributes', 'attribute_dict', 'group_dict', 'group_counter')

    def __init__(self, attributes: Iterable = None):
        self.attributes = []

        # Internal index attributes
        self.attribute_dict = {}
        self.group_dict = {}
        self.group_counter = 1
        if attributes is not None:
            self._from_iterable(attributes)

    def get_next_group_identifier(self) -> str:
        """
        Retrieve the next un-used attribute group identifier
        and increment the internal counter.

        Returns
        -------
        str
        """
        next_value = self.group_counter
        self.group_counter += 1
        return str(next_value)

    def _add_attribute_raw(self, attr: Attribute):
        key = attr.key
        group_identifier = attr.group_id
        if group_identifier is not None:
            int_group_identifier = int(group_identifier)
            if int_group_identifier <= self.group_counter:
                self.group_counter = int_group_identifier + 1
            attr.group_id = group_identifier

        self.attributes.append(attr)
        index = len(self.attributes) - 1

        #### If there is already one of these, add it to the lists in attribute_dict
        if key in self.attribute_dict:
            self.attribute_dict[key]["indexes"].append(index)
            if group_identifier is not None:
                self.attribute_dict[key]["groups"].append(group_identifier)

        #### Otherwise, create the entry in attribute_dict
        else:
            if group_identifier is not None:
                self.attribute_dict[key] = {"indexes": [
                    index], "groups": [group_identifier]}
            else:
                self.attribute_dict[key] = {"indexes": [index], "groups": []}

        #### If there is a group_identifier, then update the group_dict
        if group_identifier is not None:
            #### If this group already has one or more entries, add to it
            if group_identifier in self.group_dict:
                self.group_dict[group_identifier].append(index)
            #### Else create and entry for the group_identifier
            else:
                self.group_dict[group_identifier] = [index]

    #### Add an attribute to the list and update the lookup tables
    def add_attribute(self, key: str, value, group_identifier: Optional[str] = None, owner_id: Optional[Any] = None):
        """
        Add an attribute to the list and update the lookup tables

        Parameters
        ----------
        key : str
            The name of the attribute to add
        value : object
            The value of the attribute to add
        group_identifier : str, optional
            The attribute group identifier to use, if any. If not provided,
            no group is assumed.
        """
        if group_identifier is not None:
            int_group_identifier = int(group_identifier)
            if int_group_identifier <= self.group_counter:
                self.group_counter = int_group_identifier + 1
        attr = Attribute(key, value, group_identifier, owner_id=owner_id)
        self._add_attribute_raw(attr)

    def add_attribute_group(self, attributes: List[Union[Attribute, Tuple[str, Any]]], owner_id: Optional[Any]=None):
        """Add a collection of connected attributes that are part of a single group"""
        group_id = self.get_next_group_identifier()
        for attr in attributes:
            if isinstance(attr, Attribute):
                key = attr.key
                value = attr.value
            else:
                key, value = attr
            self.add_attribute(key, value, group_id, owner_id)

    def get_attribute(self, key: str, group_identifier: Optional[str] = None,
                      raw: bool = False) -> Union[Any, List[Any], Attribute,
                                                  List[Attribute]]:
        """
        Get the value or values associated with a given
        attribute key.

        Parameters
        ----------
        key : str
            The name of the attribute to retrieve
        group_identifier : str, optional
            The specific group identifier to return from.
        raw : bool
            Whether to return the :class:`Attribute` object or unwrap the value

        Returns
        -------
        attribute_value: object or list[object]
            Returns single or multiple values for the requested attribute.
        """
        indices_and_groups = self.attribute_dict[key]
        if group_identifier is None:
            indices = indices_and_groups['indexes']
            if len(indices) > 1:
                if raw:
                    return [self.attributes[i] for i in indices]
                return [self.attributes[i][1] for i in indices]
            else:
                if raw:
                    return self.attributes[indices[0]]
                return self.attributes[indices[0]][1]
        else:
            groups = indices_and_groups['groups']
            try:
                i = groups.index(group_identifier)
            except ValueError:
                raise KeyError(f"Group {group_identifier} not found") from None
            indices = indices_and_groups['indexes']
            idx = indices[i]
            if raw:
                return self.attributes[idx]
            return self.attributes[idx][1]

    def get_attribute_group(self, group_identifier: str) -> List[Any]:
        """Get all the members of a specified attribute group"""
        result = []
        group_identifier = str(group_identifier)
        for k, indices_and_groups in self.attribute_dict.items():
            for i, g in zip(indices_and_groups['indexes'], indices_and_groups['groups']):
                if g == group_identifier:
                    result.append(self.attributes[i])
        return result

    def replace_attribute(self, key, value, group_identifier=None, owner_id=None):
        try:
            indices_and_groups = self.attribute_dict[key]
        except KeyError:
            return self.add_attribute(key, value, group_identifier=group_identifier, owner_id=owner_id)
        if group_identifier is None:
            indices = indices_and_groups['indexes']
            if len(indices) > 1:
                raise ValueError("Cannot replace the value of an attribute used multiple times")
            else:
                self.attributes[indices[0]].value = value
        else:
            raise NotImplementedError()

    def get_by_name(self, name: str):
        """
        Search for an attribute by human-readable name.

        Parameters
        ----------
        name: str
            The name to search for.

        Returns
        -------
        object:
            The attribute value if found or :const:`None`.
        """
        matches = []
        for attr in self:
            if attr.key.split("|")[-1] == name:
                matches.append(attr[1])
        n = len(matches)
        if n == 1:
            return matches[0]
        elif n > 1:
            return matches
        return None

    def clear(self):
        """Remove all content from the store."""
        self._clear_attributes()

    def remove_attribute(self, key, group_identifier=None):
        """
        Remove the value or values associated with a given
        attribute key from the store.

        This rebuilds the entire store, which may be expensive.

        Parameters
        ----------
        key : str
            The name of the attribute to retrieve
        group_identifier : str, optional
            The specific group identifier to return from.

        """
        indices_and_groups = self.attribute_dict[key]
        if group_identifier is None:
            indices = indices_and_groups['indexes']
            if len(indices) > 1:
                indices = sorted(indices, reverse=True)
                for i in indices:
                    self.attributes.pop(i)
            else:
                self.attributes.pop(indices[0])
        else:
            groups = indices_and_groups['groups']
            i = groups.index(group_identifier)
            indices = indices_and_groups['indexes']
            idx = indices[i]
            self.attributes.pop(idx)
        attributes = self.attributes
        self.clear()
        self._from_iterable(attributes)

    def _remove_attribute_group(self, group_identifier):
        group_indices = self.group_dict.pop(group_identifier)
        for offset, i in enumerate(sorted(group_indices)):
            self.attributes.pop(i - offset)
        attributes = self.attributes
        self.clear()
        self._from_iterable(attributes)

    def _iter_attribute_groups(self):
        seen = set()
        for group_id, indices in self.group_dict.items():
            yield group_id, [self.attributes[i] for i in indices]
            seen.update(indices)
        acc = []
        for i, attr in enumerate(self.attributes):
            if i in seen:
                continue
            acc.append(attr)
        yield None, acc

    def has_attribute(self, key):
        """
        Test for the presence of a given attribute

        Parameters
        ----------
        key : str
            The attribute to test for

        Returns
        -------
        bool
        """
        return key in self.attribute_dict

    def __contains__(self, key):
        return self.has_attribute(key)

    def __getitem__(self, key):
        return self.get_attribute(key)

    def __setitem__(self, key, value):
        if self.has_attribute(key):
            self.replace_attribute(key, value)
        else:
            self.add_attribute(key, value)

    def keys(self):
        return self.attribute_dict.keys()

    def __eq__(self, other):
        if other is None:
            return False
        self_keys = self.keys()
        other_keys = other.keys()
        if len(self_keys) != len(other_keys):
            return False
        for key in self_keys:
            try:
                self_value = self.get_attribute(key)
                other_value = other.get_attribute(key)
            except KeyError:
                return False
            if self_value != other_value:
                return False
        return True

    def __ne__(self, other):
        return not self == other

    def __len__(self):
        return self._count_attributes()

    def __bool__(self):
        return len(self) > 0

    def __iter__(self):
        return self._iter_attributes()

    def _count_attributes(self) -> int:
        return len(self.attributes)

    def _iter_attributes(self) -> Iterator[Attribute]:
        return iter(self.attributes)

    def _clear_attributes(self):
        self.attributes = []

        # Internal index attributes
        self.attribute_dict = {}
        self.group_dict = {}
        self.group_counter = 1

    def _from_iterable(self, attributes):
        mapping = {}
        for attrib in attributes:
            if not isinstance(attrib, Attribute):
                attrib = Attribute(*attrib)
            else:
                attrib = attrib.copy()
            if attrib.group_id is not None:
                # this is a grouped attribute
                if attrib.group_id in mapping:
                    remap = mapping[attrib.group_id]
                else:
                    remap = self.get_next_group_identifier()
                    mapping[attrib.group_id] = remap
                attrib.group_id = remap
                self._add_attribute_raw(attrib)
            else:
                self._add_attribute_raw(attrib)

    def _attributes_from_iterable(self, attributes):
        return self._from_iterable(attributes)

    def copy(self):
        """Make a deep copy of the object"""
        return self.__class__(self.attributes)

    @property
    def attribute_sets(self):
        if not self.has_attribute(const.ATTRIBUTE_SET_NAME):
            return []
        attr_sets = self.get_attribute(const.ATTRIBUTE_SET_NAME)
        if not isinstance(attr_sets, list):
            attr_sets = [attr_sets]
        return attr_sets

    def __repr__(self):
        if len(self) == 0:
            return f"{self.__class__.__name__}([])"
        lines = list(map(str, self.attributes))
        template = "{}([\n{}])"
        return template.format(
            self.__class__.__name__,
            textwrap.indent(',\n'.join(lines), ' ' * 2))


class IdentifiedAttributeManager(AttributeManager):
    """
    An :class:`AttributeManager` with an :attr:`id` attribute.

    This serves as the base type for most intermediate layers in
    :title-reference:`mzSpecLib`.
    """

    __slots__ = ('id', )

    id: str

    def __init__(self, id, attributes: Iterable = None):
        self.id = str(id)
        super(IdentifiedAttributeManager, self).__init__(attributes)

    def __repr__(self):
        template = f"{self.__class__.__name__}(id={self.id}, "
        lines = list(map(str, self.attributes))
        if not lines:
            template += "[])"
            return template
        template += "[\n%s])" % textwrap.indent(',\n'.join(lines), ' ' * 2)
        return template


class _ReadAttributes(object):
    """
    A mixin type that provides read access to controlled vocabulary-backed attributes.

    Assumes that the implementation class will provide an :attr:`attributes` attribute
    that is an instance of :class:`AttributeManager`, or at least :class:`Attributed`.
    """

    __slots__ = ()

    attributes: AttributeManager

    def get_attribute(self, key, group_identifier=None, raw: bool = False):
        """
        Get the value or values associated with a given
        attribute key from the entity's attribute store.

        Parameters
        ----------
        key : str
            The name of the attribute to retrieve
        group_identifier : str, optional
            The specific group identifier to return from.
        raw : bool, optional
            To return the stored value, or an :class:`Attribute` object preserving
            additional information

        Returns
        -------
        attribute_value: object or list[object]
            Returns single or multiple values for the requested attribute.
        """
        return self.attributes.get_attribute(key, group_identifier=group_identifier, raw=raw)

    def get_attribute_group(self, group_identifier: str) -> List[Any]:
        return self.attributes.get_attribute_group(group_identifier)

    def has_attribute(self, key) -> bool:
        """
        Test for the presence of a given attribute in the library
        level store.

        Parameters
        ----------
        key : str
            The attribute to test for

        Returns
        -------
        bool
        """
        return self.attributes.has_attribute(key)

    def get_by_name(self, name: str):
        """
        Search for an attribute by human-readable name.

        Parameters
        ----------
        name: str
            The name to search for.

        Returns
        -------
        object:
            The attribute value if found or :const:`None`.
        """
        return self.attributes.get_by_name(name)

    def _iter_attribute_groups(self):
        return self.attributes._iter_attribute_groups()

    def _count_attributes(self) -> int:
        return self.attributes._count_attributes()

    def _iter_attributes(self) -> Iterator[Attribute]:
        return self.attributes._iter_attributes()

    @property
    def attribute_sets(self):
        return self.attributes.attribute_sets


class _WriteAttributes(object):
    """
    A mixin type that provides write access to controlled vocabulary-backed attributes.

    Assumes that the implementation class will provide an :attr:`attributes` attribute
    that is an instance of :class:`AttributeManager`, or at least :class:`Attributed`.
    """

    __slots__ = ()

    attributes: AttributeManager

    def add_attribute(self, key, value, group_identifier=None) -> Union[Any, List[Any]]:
        """
        Add an attribute to the entity's attributes store.

        Parameters
        ----------
        key : str
            The name of the attribute to add
        value : object
            The value of the attribute to add
        group_identifier : str, optional
            The attribute group identifier to use, if any. If not provided,
            no group is assumed.
        """
        return self.attributes.add_attribute(key, value, group_identifier=group_identifier)

    def add_attribute_group(self, attributes: List[Union[Attribute, Tuple[str, Any]]], owner_id: Optional[Any]):
        """Add a collection of connected attributes that are part of a single group"""
        self.attributes.add_attribute_group(attributes, owner_id=owner_id)

    def get_next_group_identifier(self) -> str:
        """
        Retrieve the next un-used attribute group identifier
        and increment the internal counter.

        Returns
        -------
        str
        """
        return self.attributes.get_next_group_identifier()

    def replace_attribute(self, key, value, group_identifier=None):
        return self.attributes.replace_attribute(key, value, group_identifier=group_identifier)

    def remove_attribute(self, key, group_identifier=None):
        """
        Remove the value or values associated with a given
        attribute key from the entity's attribute store.

        This rebuilds the entire store, which may be expensive.

        Parameters
        ----------
        key : str
            The name of the attribute to retrieve
        group_identifier : str, optional
            The specific group identifier to return from.

        """
        return self.attributes.remove_attribute(key, group_identifier=group_identifier)

    def _attributes_from_iterable(self, attributes):
        return self.attributes._attributes_from_iterable(attributes)

    def _clear_attributes(self):
        return self.attributes._clear_attributes()


class AttributedEntity(_ReadAttributes, _WriteAttributes):
    """
    A base type for entities which contain an :class:`AttributeManager`
    without being completely subsumed by it.

    An :class:`AttributeManager` represents a collection of attributes
    first and foremost, supplying :class:`~.collections.abc.MutableMapping`-like
    interface to them, in addition to methods.
    """

    __slots__ = ("attributes", )

    attributes: AttributeManager

    def __init__(self, attributes: Iterable=None, **kwargs):
        self.attributes = AttributeManager(attributes)
        super().__init__(**kwargs)


Attributed = Union[AttributeManager, AttributedEntity]


class AttributeManagedProperty(Generic[T]):
    """
    A :class:`property`-like object that stores its value as a controlled vocabulary backed
    attribute on an :class:`Attributed`-like object.

    Attributes
    ----------
    attribute : str
        The attribute which is used to read/write the value of the descriptor.
    multiple : bool
        Whether the default assumption is that the attribute will be repeated.
        When :const:`True`, a :class:`list` is always returned.
    """

    __slots__ = ("attribute", "multiple")
    attribute: str
    multiple: bool

    def __init__(self, attribute: str, multiple: bool = False):
        self.attribute = attribute
        self.multiple = multiple

    def _get_group_id(self, inst: AttributeManager) -> Optional[str]:
        return getattr(inst, "group_id", None)

    def __get__(self, inst: AttributeManager, cls: Type) -> T:
        if inst is None:
            return self
        if inst.has_attribute(self.attribute):
            value = inst.get_attribute(self.attribute, group_identifier=self._get_group_id(inst))
            if self.multiple and not isinstance(value, list):
                value = [value]
            return value
        return None

    def __set__(self, inst: AttributeManager, value: T):
        attrib = self.attribute
        if inst.has_attribute(attrib):
            inst.replace_attribute(attrib, value, group_identifier=self._get_group_id(inst))
        elif inst._count_attributes() > 0:
            attribs = [[attrib, value]] + list(inst._iter_attributes())
            inst._clear_attributes()
            inst._attributes_from_iterable(attribs)
        else:
            inst.add_attribute(attrib, value, group_identifier=self._get_group_id(inst))

    def __delete__(self, inst: AttributeManager, attr):
        inst.remove_attribute(self.attribute, group_identifier=self._get_group_id(inst))


class AttributeListManagedProperty(Generic[T]):
    """
    Like :class:`AttributeManagedProperty`, except a list of attributes
    are tried in succession until one is found.

    Attributes
    ----------
    attributes : list[str]
        The attributes to probe, in order
    """

    __slots__ = ("attributes", )
    attributes: List[str]

    def __init__(self, attributes: List[str]):
        self.attributes = attributes

    def __get__(self, inst: AttributeManager, cls: Type) -> T:
        if inst is None:
            return self
        key, val = self._find_key_used(inst)
        return val

    def _find_key_used(self, inst: AttributeManager) -> Optional[Tuple[str, T]]:
        for attr in self.attributes:
            try:
                return attr, inst.get_attribute(attr)
            except KeyError:
                continue
        return None, None

    def __set__(self, inst: AttributeManager, value: T):
        key, _val = self._find_key_used(inst)
        if key is None:
            attrib = self.attributes[0]
        else:
            attrib = key
        if inst.has_attribute(attrib):
            inst.replace_attribute(attrib, value)
        elif inst._count_attributes() > 0:
            attribs = [[attrib, value]] + list(inst._iter_attributes())
            inst._clear_attributes()
            inst._attributes_from_iterable(attribs)
        else:
            inst.add_attribute(attrib, value)

    def __delete__(self, inst: AttributeManager, attr):
        key, val = self._find_key_used(inst)
        if key is not None:
            inst.remove_attribute(key)


class AttributeProxyMeta(type):
    """
    A metaclass that manages controlled vocabulary-backed properties
    like :class:`AttributeManagedProperty` or :class:`AttributeListManagedProperty`.

    This includes generating a meaningful :meth:`__repr__` method for the instance
    types.
    """

    def __new__(cls, typename, bases, namespace):
        props = {}
        for k, v in namespace.items():
            if isinstance(v, (AttributeManagedProperty, AttributeListManagedProperty)):
                props[k] = v
        for base in bases[::-1]:
            if hasattr(base, '_attribute_props'):
                for k, v in base._attribute_props.items():
                    if k not in props:
                        props[k] = v
        namespace['_attribute_props'] = props
        if "__repr__" not in namespace:
            def __repr__(self):
                pairs = []
                for k in self._attribute_props:
                    try:
                        value = getattr(self, k)
                    except KeyError:
                        value = None
                    pairs.append((k, value))
                content = ', '.join(f"{k}={v!r}" for k, v in pairs)
                return f"{self.__class__.__name__}({content})"
            namespace['__repr__'] = __repr__
        return super().__new__(cls, typename, bases, namespace)


class ROAttributeProxy(_ReadAttributes, metaclass=AttributeProxyMeta):
    """
    A read-only attribute proxy wraps another :class:`Attributed`-like
    object and provides tailored interactions with all or a subset
    of those attributes.
    """

    attributes: Attributed

    def __init__(self, attributes):
        self.attributes = attributes


AttributeProxy = ROAttributeProxy


class AttributeSet(AttributedEntity):
    name: str
    _names_to_override: Set[str]

    def __init__(self, name: str, attributes: Iterable = None, **kwargs):
        super().__init__(attributes, **kwargs)
        self.name = name
        self._names_to_override = self._get_names_to_override()

    def _get_names_to_override(self):
        keys = set()
        for attr in self.attributes:
            if attr.group_id is None:
                keys.add(attr.key)
        return keys

    def member_of(self, target: Attributed) -> bool:
        for attrib in self.attributes:
            if attrib.group_id:
                raise NotImplementedError()
            if not target.has_attribute(attrib.key):
                return False
        return True

    def remove(self, target: Attributed, group_identifier: Optional[str]=None):
        for attr in self.attributes:
            target.remove_attribute(attr.key, attr.value, group_identifier)

    def apply(self, target: Attributed, group_identifier: Optional[str]=None):
        if group_identifier is not None:
            for attr in self.attributes:
                target.add_attribute(attr.key, attr.value, group_identifier, owner_id=self.name)
        else:
            terms_to_remove: List[Tuple[str, Union[Attribute, List[Attribute]]]] = []
            for key in self.attributes.keys():
                if target.has_attribute(key):
                    terms_to_remove.append((key, target.get_attribute(key, raw=True)))

            group_ids = DefaultDict(int)
            for key, terms in terms_to_remove:
                if isinstance(terms, list):
                    for term in terms:
                        if term.group_id:
                            group_ids[term.group_id] += 1
                        target.remove_attribute(key, group_identifier=term.group_id)
                else:
                    term = terms
                    if term.group_id:
                        group_ids[term.group_id] += 1
                    target.remove_attribute(key, group_identifier=term.group_id)

            for group_identifier in group_ids:
                target._remove_attribute_group(group_identifier)

            for group_identifier, attrs in self._iter_attribute_groups():
                if group_identifier is None:
                    for a in attrs:
                        target.add_attribute(a.key, a.value, group_identifier=None, owner_id=self.name)
                else:
                    target.add_attribute_group(attrs)

    def __repr__(self):
        template = f"{self.__class__.__name__}(name={self.name}, "
        lines = list(map(str, self.attributes))
        if not lines:
            template += "[])"
            return template
        template += "[\n%s])" % textwrap.indent(',\n'.join(lines), ' ' * 2)
        return template


@dataclass
class AttributeSetRef:
    attribute_set: AttributeSet
    group_id: str

    @property
    def name(self):
        return self.attribute_set.name

    def apply(self, target: Attributed, group_identifier: Optional[str] = None):
        if group_identifier is None:
            group_identifier = self.group_id
        return self.attribute_set.apply(target, group_identifier)

    def remove(self, target: Attributed, group_identifier: Optional[str]=None):
        if group_identifier is None:
            group_identifier = self.group_id
        return self.attribute_set.remove(target, group_identifier)


class AttributeFacet(Generic[T]):
    """
    A descriptor that resembles an object, backed by one or more attributes
    from a :class:`AttributeManager`-like type.
    """

    facet_type: Type[T]

    def __init__(self, facet_type: Type[T]):
        self.facet_type = facet_type

    def __get__(self, inst: Attributed, klass: Type[Attributed]) -> T:
        if inst is not None:
            return self.facet_type(inst)
        else:
            return self


class AttributeGroupFacet(Generic[T]):
    facet_type: Type[T]

    def __init__(self, facet_type: Type[T]):
        self.facet_type = facet_type

    def wraps(self, attributed: Attributed) -> List[T]:
        attribute_groups = []
        facet_keys = {a.attribute for a in self.facet_type._attribute_props.values()}
        for group_id in attributed.group_dict:
            group = attributed.get_attribute_group(group_id)
            group_keys = {a.key for a in group}
            if group_keys & facet_keys:
                attribute_groups.append(group_id)

        facets = []
        for group_id in attribute_groups:
            facet = self.facet_type(attributed)
            facet.group_id = group_id
            facets.append(facet)
        return facets

    def __get__(self, inst: Attributed, klass) -> Union[List[T], 'AttributeGroupFacet']:
        if inst is not None:
            return self.wraps(inst)
        return self




"""
The components of the semantic validator for :title-reference:`MzSpecLib`

To obtain the base validator, use :func:`~.load_default_validator`, and add
successive validation rule sets using :func:`~.get_validator_for` and add
them to the base rules using the :meth:`~.ValidatorBase.chain` or the `|=`
operator.

.. code-block:: python

    from mzspeclib.validate import *

    chain = load_default_validator()
    chain.validate_library(library)

    by_level: DefaultDict[RequirementLevel, List[ValidationError]] = DefaultDict(list)
    for message in chain.error_log:
        by_level[message.requirement_level].append(message)

    for level, bucket in sorted(by_level.items()):
        log_level = logging.WARN
        if level == RequirementLevel.may:
            log_level = logging.DEBUG
        logger.log(log_level, f"Found {len(bucket)} violations for {level.name.upper()} rules")
        for err in bucket:
            logger.log(log_level, f"... {err.message}")
"""

from .level import CombinationLogic, RequirementLevel
from .semantic_rule import (RuleSet, ScopedSemanticRule, AttributeSemanticPredicate, AttributeSemanticRule)
from .object_rule import (LibraryFormatVersionFirstRule, SpectrumPeakAnnotationRule, ScopedObjectRuleBase)
from .validator import (
    ValidationError, ValidationWarning, ControlledVocabularyAttributeValidator, ValidatorBase,
    RuleValidator, ValidatorChain, load_default_validator, get_object_validator_for, get_validator_for
)


"""Configuration file for the Sphinx documentation builder."""

# Project information
project = "mzSpecLib"
author = "HUPO-PSI"
copyright = "Creative Commons Attribution 4.0 International"
github_project_url = "https://github.com/HUPO-PSI/mzspeclib-py"
github_doc_root = "https://github.com/HUPO-PSI/mzspeclib-py/tree/main/docs"

# General configuration
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinx_click.ext",
    "myst_parser",
]
source_suffix = [".rst"]
master_doc = "index"
exclude_patterns = ["_build"]

# Options for HTML output
html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]
html_css_files = ["css/custom.css"]

# Autodoc options
autodoc_default_options = {"members": True, "show-inheritance": True}
autodoc_member_order = "bysource"
autodoc_typehints = "description"
autoclass_content = "init"

# Intersphinx options
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "psims": ("https://mobiusklein.github.io/psims/docs/build/html/", None),
    "pyteomics": ("https://pyteomics.readthedocs.io/en/stable/", None),
}


def setup(app):
    """Set up the Sphinx application."""
    config = {"enable_eval_rst": True}  # noqa: F841



import io
import os
import numbers
import pathlib
import logging
from typing import Iterator, List, Union

from sqlalchemy import Column, ForeignKey, Integer, Float, String, DateTime, Text, LargeBinary

from mzspeclib.index.base import IndexRecordBase
try: # For SQLAlchemy 2.0
    from sqlalchemy.orm import declarative_base
except ImportError:
    from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy.engine import Engine
from sqlalchemy import create_engine, func
from sqlalchemy.orm import sessionmaker, scoped_session

from .base import IndexBase, IndexInitializedRecord


logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


Base = declarative_base()

#### Define the database tables as classes


class SpectrumLibraryIndexAttribute(Base):
    __tablename__ = 'spectrum_library_index_attribute'
    id = Column(Integer, primary_key=True)
    name = Column(String(255), nullable=False)
    value = Column(String(1024), nullable=False)


#### Define the database tables as classes
class SpectrumLibraryIndexRecord(Base):
    __tablename__ = 'spectrum_library_index_record'
    id = Column(Integer, primary_key=True)
    number = Column(Integer, nullable=False, index=True)
    offset = Column(Integer, nullable=False)
    name = Column(String(1024), nullable=False)
    index = Column(Integer, nullable=False, index=True)
    analyte = Column(String(2014), nullable=True)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.number}, {self.offset}, {self.name}, {self.analyte})"


class ClusterSpectrumLibraryIndexRecord(Base):
    __tablename__ = 'cluster_spectrum_library_index_record'
    id = Column(Integer, primary_key=True)
    number = Column(Integer, nullable=False, index=True)
    offset = Column(Integer, nullable=False)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.number}, {self.offset}, {self.name}, {self.analyte})"


class SQLIndex(IndexBase):
    """
    An on-disk data structure for holding spectrum metadata and offsets.

    This uses a SQLite3 database with the file extension ``.splindex`` to
    hold information.

    Attributes
    ----------
    session : :class:`~.sqlalchemy.orm.scoped_session`
        A thread-aware database session manager
    """

    extension = '.splindex'

    filename: str
    index_filename: str
    _cache: SpectrumLibraryIndexRecord
    session: scoped_session
    engine: Engine

    @classmethod
    def from_filename(cls, filename, library=None):
        if not isinstance(filename, (str, pathlib.Path)):
            if not hasattr(filename, "name"):
                raise TypeError(f"Could not coerce filename from {filename}")
            else:
                filename = filename.name
        exists = os.path.exists(filename + cls.extension)
        inst = cls(filename)

        # File was empty
        if len(inst) == 0 and exists:
            exists = False
        return IndexInitializedRecord(inst, exists)

    @classmethod
    def exists(cls, filename: Union[str, pathlib.Path, io.FileIO]):
        """Check if an index file exists"""
        if not isinstance(filename, (str, pathlib.Path)):
            if not hasattr(filename, "name"):
                raise TypeError(f"Could not coerce filename from {filename}")
            else:
                filename = filename.name
        exists = os.path.exists(filename + cls.extension)
        return exists

    def __init__(self, filename):
        self.filename = filename
        self.index_filename = self.filename + self.extension
        self._cache = None
        self._connect()
        self._size = len(self)
        self._size_uncommitted = 0

    def _connect(self, create=None):
        filename = self.index_filename
        if os.path.exists(filename):
            if create:
                logger.debug(f'INFO: Deleting previous index file {filename}')
                os.remove(filename)
        logger.debug(f'INFO: Creating index file {filename}')
        engine = create_engine("sqlite:///"+filename)
        Base.metadata.create_all(engine)

        session = scoped_session(sessionmaker(bind=engine))
        self.session = session
        self.engine = engine
        self._cache = None

    def add(self, number, offset, name, analyte=None, attributes=None):
        """
        Add a new entry to the spectrum index.

        Parameters
        ----------
        number : int
            A numerical identifier for this spectrum.
        offset : int
            The offset in the file to reach the spectrum (in bytes if appropriate)
        name : str,
            A text identifier for this spectrum.
        analyte : str, optional
            A text representation of the analyte for that record
        attributes : Dict[str, Any], optional
            A key-value pair collection of this record, currently not supported.
        """
        record = SpectrumLibraryIndexRecord(number=number, offset=offset, name=name,
                                            index=self._size + self._size_uncommitted, analyte=analyte)
        self._size_uncommitted += 1
        if attributes is not None:
            raise NotImplementedError("Record attribute storage is not implemented")
        self.session.add(record)

    def add_cluster(self, number: int, offset: int, attributes=None):
        """
        Add a new entry to the spectrum index.

        Parameters
        ----------
        number : int
            A numerical identifier for this spectrum.
        offset : int
            The offset in the file to reach the spectrum (in bytes if appropriate)
        attributes : Dict[str, Any], optional
            A key-value pair collection of this record, currently not supported.
        """
        record = ClusterSpectrumLibraryIndexRecord(number=number, offset=offset)
        if attributes is not None:
            raise NotImplementedError("Record attribute storage is not implemented")
        self.session.add(record)

    def commit(self):
        """Persist any new entries to disk."""
        self._size += self._size_uncommitted
        self._size_uncommitted = 0
        self.session.commit()

    def iter_clusters(self) -> Iterator[IndexRecordBase]:
        """Iterate over cluster entries in the index."""
        for record in self.session.query(ClusterSpectrumLibraryIndexRecord).order_by(
                ClusterSpectrumLibraryIndexRecord.number).yield_per(10000):
            yield record

    def iter_spectra(self):
        """Iterate over spectrum entries in the index."""
        for record in self.session.query(SpectrumLibraryIndexRecord).order_by(
                SpectrumLibraryIndexRecord.number).yield_per(10000):
            yield record

    def __getitem__(self, i):
        return self._get_by_index(i)

    def _get_by_index(self, i: Union[int, slice]) -> Union[SpectrumLibraryIndexRecord, List[SpectrumLibraryIndexRecord]]:
        if isinstance(i, slice):
            records = self.session.query(SpectrumLibraryIndexRecord).slice(i.start, i.stop).all()
            if i.step:
                raise NotImplementedError()
            return records
        else:
            record = self.session.query(SpectrumLibraryIndexRecord).offset(i).limit(1).first()
            return record

    def __len__(self):
        value = self.session.query(func.count(SpectrumLibraryIndexRecord.id)).scalar()
        return value

    def search(self, i, **kwargs):
        if i is None and kwargs:
            # Executing attribute query
            raise NotImplementedError()
        if isinstance(i, numbers.Integral):
            if i < 0:
                i = len(self) + i
            if self._cache is not None and self._cache.number == i:
                return self._cache
            records = self.session.query(SpectrumLibraryIndexRecord).filter(SpectrumLibraryIndexRecord.number == i).all()

            if len(records) == 1:
                return records[0]
            elif len(records) == 0:
                raise IndexError(i)
            else:
                raise ValueError(f"Too many records found for spectrum number {i}")
        elif isinstance(i, slice):
            start = i.start or 0
            end = i.stop or float('inf')
            records = self.session.query(SpectrumLibraryIndexRecord).filter(
                SpectrumLibraryIndexRecord.number >= start,
                SpectrumLibraryIndexRecord.number < end).all()
            return records
        else:
            records = self.session.query(SpectrumLibraryIndexRecord).filter(
                SpectrumLibraryIndexRecord.name == i).all()
            if not records:
                raise KeyError(i)
            elif len(records) == 1:
                return records[0]
            else:
                return records

    def search_clusters(self, i, **kwargs):
        if i is None and kwargs:
            # Executing attribute query
            raise NotImplementedError()
        if isinstance(i, numbers.Integral):
            if i < 0:
                i = len(self) + i
            if self._cache is not None and self._cache.number == i:
                return self._cache
            records = self.session.query(ClusterSpectrumLibraryIndexRecord).filter(
                ClusterSpectrumLibraryIndexRecord.number == i).all()

            if len(records) == 1:
                return records[0]
            elif len(records) == 0:
                raise IndexError(i)
            else:
                raise ValueError(f"Too many records found for spectrum number {i}")
        elif isinstance(i, slice):
            start = i.start or 0
            end = i.stop or float('inf')
            records = self.session.query(ClusterSpectrumLibraryIndexRecord).filter(
                ClusterSpectrumLibraryIndexRecord.number >= start,
                ClusterSpectrumLibraryIndexRecord.number < end).all()
            return records
        else:
            raise NotImplementedError()



import warnings

from typing import Iterable, KeysView, ItemsView, Optional, ValuesView, Dict, MutableMapping, Mapping

from pyteomics import proforma

from mzspeclib.attributes import AttributedEntity, IdentifiedAttributeManager, AttributeManagedProperty, AttributeProxy, AttributeGroupFacet
from mzspeclib.const import (ANALYTE_MIXTURE, CHARGE_STATE, PROFORMA_ION, PROFORMA_SEQ, STRIPPED_PEPTIDE_SEQ, FIRST_ANALYTE_KEY, FIRST_INTERPRETATION_KEY)


class _AnalyteMappingProxy(Mapping[str, 'Analyte']):
    parent: Mapping[str, 'Interpretation']

    def __init__(self, parent):
        self.parent = parent

    def __getitem__(self, key):
        for group_id, group in self.parent.items():
            if key in group:
                return group[key]
        raise KeyError(key)

    def __iter__(self):
        keys = set()
        for group_id, group in self.parent.items():
            k_ = set(group.keys())
            keys.update(k_)
        return iter(keys)

    def __len__(self):
        return sum([len(v) for v in self.parent.values()])

    def __repr__(self):
        d = dict(self)
        return f"{self.__class__.__name__}({d})"


class InterpretationCollection(MutableMapping[str, 'Interpretation']):
    """
    A mutable mapping for :class:`Interpretation`s that also exposes a shared pool of
    :class:`Analyte` members.
    """

    __slots__ = ('interpretations', )

    interpretations: Dict[str, 'Interpretation']

    def __init__(self, interpretations=None):
        if interpretations is None:
            interpretations = {}
        self.interpretations = interpretations

    def get_interpretation(self, interpretation_id) -> 'Interpretation':
        """Get the interpretation given by the key ``interpretation_id``"""
        return self.interpretations[str(interpretation_id)]

    def add_interpretation(self, interpretation: 'Interpretation'):
        """
        Add an :class:`Interpretation` to the collection.

        The key used will be :attr:`Interpretation.id`.
        """
        self.set_interpretation(str(interpretation.id), interpretation)

    def set_interpretation(self, key, interpretation: 'Interpretation'):
        """Set the interpretation for ``key`` to ``interpretation``"""
        self.interpretations[str(key)] = interpretation

    def __getitem__(self, key) -> 'Interpretation':
        return self.get_interpretation(key)

    def __setitem__(self, key, value: 'Interpretation'):
        self.set_interpretation(key, value)

    def __delitem__(self, key):
        del self.interpretations[str(key)]

    def __len__(self):
        return len(self.interpretations)

    def __contains__(self, key):
        return key in self.interpretations

    def __iter__(self):
        return iter(self.interpretations)

    def keys(self) -> KeysView[str]:
        return self.interpretations.keys()

    def values(self) -> ValuesView['Interpretation']:
        return self.interpretations.values()

    def items(self) -> ItemsView[str, 'Interpretation']:
        return self.interpretations.items()

    @property
    def analytes(self) -> '_AnalyteMappingProxy':
        """A facade that exposes a :class:`Mapping[str, Analyte]` interface"""
        return _AnalyteMappingProxy(self)

    def __repr__(self):
        d = dict(self)
        return f"{self.__class__.__name__}({d})"


class Interpretation(AttributedEntity, MutableMapping):
    """
    An interpretation of a :class:`~.Spectrum` with one or more :class:`~.Analyte`
    members.

    Attributes
    ----------
    id : str
        The identifier for the interpretation
    analytes : dict[str, :class:`Analyte`]
        The analytes which are part of interpretations.
    member_interpretations : dict[str, :class:`InterpretationMember`]
        The interpretation details which are associated with specific :class:`Analyte` members.
    """

    __slots__ = ('id', 'analytes', 'member_interpretations')

    id: str
    analytes: Dict[str, 'Analyte']
    member_interpretations: Dict[str, 'InterpretationMember']

    def __init__(self, id, attributes: Iterable = None, analytes: Dict = None, member_interpretations: Dict = None):
        self.id = str(id)
        self.analytes = analytes or {}
        self.member_interpretations = member_interpretations or {}
        super(Interpretation, self).__init__(attributes)

    def _update_mixture_members_term(self):
        value = sorted(map(int, self.analytes.keys()))
        self.replace_attribute(ANALYTE_MIXTURE, value)

    def get_analyte(self, analyte_id) -> 'Analyte':
        """Retrieve an analyte by its identifier"""
        return self.analytes[str(analyte_id)]

    def add_analyte(self, analyte: 'Analyte'):
        """Add an analyte to the interpretation"""
        self.set_analyte(analyte.id, analyte)
        self._update_mixture_members_term()

    def set_analyte(self, key, analyte: 'Analyte'):
        """Set the analyte for ``key`` to ``analyte``"""
        key = str(key)
        self.analytes[key] = analyte
        self._update_mixture_members_term()

    def remove_analyte(self, analyte_id):
        """Remove the analyte for ``analyte_id``"""
        del self.analytes[str(analyte_id)]
        self._update_mixture_members_term()

    def has_analyte(self, analyte_id) -> bool:
        """Check if this interpretation includes ``analyte_id``"""
        return str(analyte_id) in self.analytes

    def get_member_interpretation(self, member_id) -> 'InterpretationMember':
        """Retrieve the :class:`InterpretationMember` for the ``member_id``"""
        return self.member_interpretations[str(member_id)]

    def add_member_interpretation(self, interpretation_member: 'InterpretationMember'):
        """Add an :class:`InterpretationMember` to the interpretation"""
        self._set_member_interpretation(interpretation_member.id, interpretation_member)

    def _set_member_interpretation(self, key, interpretation_member: 'InterpretationMember'):
        self.member_interpretations[str(key)] = interpretation_member

    def remove_member_interpretation(self, member_id):
        """Remove the :class:`InterpretationMember` for ``member_id``"""
        del self.member_interpretations[str(member_id)]

    def validate(self) -> bool:
        """
        Perform validation on each component to confirm this object is well formed.

        Returns
        -------
        bool
        """
        analyte_ids = set(self.analytes)
        member_ids = set(self.member_interpretations)
        valid = True
        if not (analyte_ids >= member_ids):
            warnings.warn(
                f"Interpretation has InterpretationMembers {member_ids - analyte_ids} lacking Analytes")
            valid = False
        return valid

    def __getitem__(self, key) -> 'Analyte':
        return self.get_analyte(key)

    def __setitem__(self, key, value: 'Analyte'):
        self.set_analyte(key, value)

    def __delitem__(self, key):
        self.remove_analyte(key)

    def __iter__(self):
        return iter(self.analytes)

    def __len__(self):
        return len(self.analytes)

    def __repr__(self):
        d = dict(self)
        a = ''
        if self.attributes:
            a = f', {self.attributes}'
        return f"{self.__class__.__name__}({d}{a})"


class InterpretationMember(IdentifiedAttributeManager):
    """
    A collection of attributes associated with a specific :class:`Analyte`
    contained in an :class:`Interpretation`
    """

    __slots__ = ()


class ProteinDescription(AttributeProxy):
    """
    A protein associated with a peptide :class:`Analyte`.

    This is a read-only proxy.
    """

    accession = AttributeManagedProperty("MS:1000885|protein accession")
    name = AttributeManagedProperty("MS:1000886|protein name")
    missed_cleavages = AttributeManagedProperty[int]("MS:1003044|number of missed cleavages")
    cleavage_agent = AttributeManagedProperty("MS:1001045|cleavage agent name")
    number_of_enzymatic_termini = AttributeManagedProperty[int]("MS:1003048|number of enzymatic termini")
    flanking_n_terminal_residue = AttributeManagedProperty("MS:1001112|n-terminal flanking residue")
    flanking_c_terminal_residue = AttributeManagedProperty("MS:1001113|c-terminal flanking residue")


class Analyte(IdentifiedAttributeManager):
    """A molecule that is associated with an :class:`Interpretation`"""

    __slots__ = ()

    mass = AttributeManagedProperty[float]("MS:1001117|theoretical mass")
    proteins = AttributeGroupFacet[ProteinDescription](ProteinDescription)

    @property
    def peptide(self) -> Optional[proforma.ProForma]:
        """
        Read out the peptide sequence of the analyte if it is present.

        This probes the following attributes, in order:
            1. MS:1003270|proforma peptidoform ion notation
            2. MS:1000889|proforma peptidoform sequence
            3. MS:1000888|stripped peptide sequence

        Returns
        -------
        :class:`~pyteomics.proforma.ProForma` or :const:`None`
        """
        if self.has_attribute(PROFORMA_SEQ):
            peptide_seq = self.get_attribute(PROFORMA_SEQ)
            return proforma.ProForma.parse(peptide_seq)
        if self.has_attribute(PROFORMA_ION):
            peptide_seq = self.get_attribute(PROFORMA_ION)
            return proforma.ProForma.parse(peptide_seq)
        if self.has_attribute(STRIPPED_PEPTIDE_SEQ):
            peptide_seq = self.get_attribute(STRIPPED_PEPTIDE_SEQ)
            return proforma.ProForma.parse(peptide_seq)
        return None

    @property
    def charge(self) -> Optional[int]:
        """
        Read the analyte's charge state, if it is present.

        This probes the following attributes in order:
            1. MS:1000041|charge state
            2. MS:1003270|proforma peptidoform ion notation

        Returns
        -------
        :class:`int` or :const:`None`
        """
        if self.has_attribute(CHARGE_STATE):
            return self.get_attribute(CHARGE_STATE)
        elif self.has_attribute(PROFORMA_ION):
            ion_val = self.get_attribute(PROFORMA_ION)
            val = proforma.ProForma.parse(ion_val)
            return val.charge_state
        else:
            return None

    @charge.setter
    def charge(self, value):
        if value is not None:
            if self.has_attribute(CHARGE_STATE):
                self.replace_attribute(CHARGE_STATE, value)
            else:
                self.add_attribute(CHARGE_STATE, value)
        else:
            self.remove_attribute(CHARGE_STATE)


#!/usr/bin/env python
import io
import tarfile
import pprint

from urllib.request import urlopen

from mzspeclib.backends import msp
from mzspeclib import annotation


urls = [
    "https://chemdata.nist.gov/download/peptide_library/libraries/human/HCD/2020_05_19/human_hcd_tryp_good.msp.tar.gz",
    "https://chemdata.nist.gov/download/peptide_library/libraries/human/HCD/2019_02_14/human_hcd_labelfree_phospho_selected_passed.msp.tar.gz",
    "https://chemdata.nist.gov/download/peptide_library/libraries/cptaclib/2015/cptac2_human_hcd_itraq_phospho_selected.msp.tar.gz",
    "https://chemdata.nist.gov/download/peptide_library/libraries/human/HCD/2022_03_16/cptac3_tmt_selected_passed_best.msp.tar.gz",
    "https://chemdata.nist.gov/download/peptide_library/libraries/sigmaups1/ION_TRAP/2011_05_24/2011_05_24_sigmaups1_consensus_final_true_lib.tar.gz",
    "https://chemdata.nist.gov/download/peptide_library/libraries/proteins/ion_trap/bsa/strict/2011_04_01/2011_04_01_bsa_consensus_final_true_lib.tar.gz",
    "https://chemdata.nist.gov/download/peptide_library/libraries/cptaclib/2015/cptac2_mouse_hcd_itraq_phospho_selected.msp.tar.gz",
]


def open_url(url: str, buffer_size: int=2 ** 20):
    buffer = io.BytesIO(
        urlopen(url).read(buffer_size)
    )

    arc = tarfile.open(
        fileobj=buffer,
        mode='r:gz')

    ti = arc.firstmember
    fh = arc.extractfile(ti)
    return fh


def check_url(url: str, n_spectra_to_read: int=100, track_unknown_attribute_values: bool=True):
    buffer = open_url(url)
    backend = msp.MSPSpectralLibrary(buffer, read_metadata=False, create_index=False)

    if track_unknown_attribute_values:
        backend.unknown_attributes = msp.UnknownKeyValueTracker()
    for i, spec in enumerate(backend.read()):
        for (mz, inten, annots, aggs) in spec.peak_list:
            for annot in annots:
                if isinstance(annot, annotation.InvalidAnnotation):
                    print(f"Failed to parse {annot} with m/z {mz}")
        if i >= n_spectra_to_read:
            break

    pprint.pprint(backend.summarize_parsing_errors())


def main():
    for url in urls:
        print(f"Testing URL {url}")
        check_url(url)


if __name__ == "__main__":
    main()



import itertools
import logging
import warnings
import re
import numbers

from dataclasses import dataclass, field
from typing import Any, Callable, Deque, Dict, Iterator, List, Optional, Sequence, Tuple, Union

from psims.controlled_vocabulary.entity import Entity, ListOfType

from mzspeclib.attributes import Attribute, Attributed

from mzspeclib.spectrum import Spectrum
from mzspeclib.analyte import Analyte, Interpretation
from mzspeclib.spectrum_library import SpectrumLibrary

from mzspeclib.ontology import _VocabularyResolverMixin


from mzspeclib.validate.level import RequirementLevel
from mzspeclib.validate.semantic_rule import ScopedSemanticRule, load_rule_set
from mzspeclib.validate.object_rule import ScopedObjectRuleBase, SpectrumPeakAnnotationRule, ValidationWarning, LibraryFormatVersionFirstRule
from mzspeclib.defaults import DEFAULT_UNITS

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def _walk_children(term: Entity):
    queue = Deque([term])
    while queue:
        term = queue.popleft()
        yield term
        queue.extend(term.children)


_is_curie = re.compile(r"(([A-Z]+):(.+))(?:\|.+)?")

def is_curie(value: str) -> bool:
    if isinstance(value, str):
        return _is_curie.match(value)
    return False


@dataclass
class ValidationContext:
    attributes_visited: Dict[Tuple[str, str], bool] = field(default_factory=dict)
    rule_states: Dict[str, Any] = field(default_factory=dict)

    def clear_attributes(self):
        self.attributes_visited.clear()

    def record_attribute(self, attribute: Union[Tuple[str, str], Attribute], result: bool):
        if isinstance(attribute, Attribute):
            attribute = (attribute.key, attribute.group_id)
        self.attributes_visited[attribute] = result

    def visited_attribute(self, attribute: Union[Tuple[str, str], Attribute]) -> bool:
        if isinstance(attribute, Attribute):
            attribute = (attribute.key, attribute.group_id)
        return attribute in self.attributes_visited



def _warning_iterator(iterator: Iterator[Spectrum]) -> Iterator[Spectrum]:
    # coerce to an actual iterator in case we were passed only an iterable
    iterator = iter(iterator)
    while True:
        try:
            with warnings.catch_warnings(record=True) as w:
                value = next(iterator)
            vw = [a for a in w if issubclass(a.category, ValidationWarning)]
            yield value, vw
        except StopIteration:
            break
        except:
            raise


def _is_of_type(attrib, relation) -> Tuple[bool, Optional[str]]:
    if isinstance(relation.value_type.type_definition, type):
        if relation.value_type.type_definition is float:
            tp = numbers.Number
        else:
            tp = relation.value_type.type_definition
        return (isinstance(attrib.value, tp), None)
    else:
        return _try_convert(attrib.value, relation.value_type.type_definition)


def _try_convert(value, converter) -> Tuple[bool, Optional[str]]:
    try:
        converter(value)
        return (True, None)
    except (ValueError, TypeError) as err:
        return (False, str(err))


class ValidatorBase(_VocabularyResolverMixin):
    name: str = None
    error_log: List
    current_context: ValidationContext

    def __init__(self, error_log=None, current_context=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.error_log = error_log or []
        self.current_context = current_context or ValidationContext()

    def reset_context(self):
        self.current_context.clear_attributes()

    def add_warning(
        self,
        obj: Attributed,
        path: str,
        identifier_path: Tuple,
        attrib: Any,
        value: Any,
        requirement_level: RequirementLevel,
        message: str,
    ):
        if hasattr(obj, "key"):
            key = obj.key
        elif hasattr(obj, "id"):
            key = obj.id
        else:
            key = ""
        warning = f"{attrib.id if hasattr(attrib, 'id') else attrib} failed to validate {path}:{identifier_path or key} ({requirement_level.name.upper()}): {message} ({self.name})"
        log_level = logging.WARN
        if requirement_level == RequirementLevel.may:
            log_level = logging.DEBUG
        logger.log(log_level, warning, stacklevel=2)
        self.error_log.append(ValidationError(path, identifier_path, attrib, value, requirement_level, warning, self.name))

    def validate_spectrum(
        self,
        spectrum: Spectrum,
        path: str,
        library: SpectrumLibrary,
        parsing_warnings: Optional[List[warnings.WarningMessage]] = None,
    ) -> bool:
        path = f"{path}/Spectrum"
        identifier_path = (spectrum.key,)
        result = self.apply_rules(spectrum, path, identifier_path)
        result &= self.check_attributes(spectrum, path, identifier_path)
        self.reset_context()

        if parsing_warnings:
            result = False
            for parsing_warning in parsing_warnings:
                logger.warn(str(parsing_warning.message))
                self.add_warning(
                    spectrum,
                    path,
                    identifier_path,
                    '',
                    RequirementLevel.must,
                    parsing_warning.message
                )

        for _key, analyte in spectrum.analytes.items():
            result &= self.validate_analyte(analyte, path, spectrum, library)

        for _key, interp in spectrum.interpretations.items():
            result &= self.validate_interpretation(interp, path, spectrum, library)
        return result

    def validate_analyte(self, analyte: Analyte, path: str, spectrum: Spectrum, library: SpectrumLibrary) -> bool:
        path = f"{path}/Analyte"
        identifier_path = (spectrum.key, analyte.id)
        result = self.apply_rules(analyte, path, identifier_path)
        result &= self.check_attributes(analyte, path, identifier_path)
        self.reset_context()
        return result

    def validate_interpretation(
        self, interpretation: Interpretation, path: str, spectrum: Spectrum, library: SpectrumLibrary
    ) -> bool:
        path = f"{path}/Interpretation"
        identifier_path = (spectrum.key, interpretation.id)
        result = self.apply_rules(interpretation, path, identifier_path)
        result &= self.check_attributes(interpretation, path, identifier_path)
        self.reset_context()
        return result

    def apply_rules(self, obj: Attributed, path: str, identifier_path: Tuple) -> bool:
        raise NotImplementedError()

    def check_attributes(self, obj: Attributed, path: str, identifier_path: Tuple) -> bool:
        """Stub implementation for any attribute rule checking"""
        return True

    def validate_library(self, library: SpectrumLibrary, spectrum_iterator: Optional[Iterator[Spectrum]]=None):
        path = "/Library"
        result = self.apply_rules(library, path, (library.identifier, ))
        result &= self.check_attributes(library, path, (library.identifier, ))
        self.reset_context()

        if spectrum_iterator is None:
            spectrum_iterator = library
        for spectrum, warns in _warning_iterator(spectrum_iterator):
            result &= self.validate_spectrum(spectrum, path, library, parsing_warnings=warns)
        return result

    def chain(self, validator: 'ValidatorBase') -> 'ValidatorBase':
        """
        Combine this validator with another validator, applying both rulesets.

        See Also
        --------
        ValidatorChain
        """
        return ValidatorChain([self, validator])

    def __or__(self, other: 'ValidatorBase') -> 'ValidatorBase':
        return self.chain(other)

    def walk_terms_for(self, curie: str) -> Iterator[str]:
        term = self.find_term_for(curie)
        for entity in _walk_children(term):
            yield f"{entity.id}|{entity.name}"


INTENSITY_UNITS = ("MS:1000131", "MS:1000132", "MS:1000905", " MS:1000814", "UO:0000269")

def _is_intensity_units(units: List):
    for unit_rel in units:
        if unit_rel.accession in INTENSITY_UNITS:
            return True
    return False


class ControlledVocabularyAttributeValidator(ValidatorBase):
    name: str = "ControlledVocabularyAttributeValidator"

    def apply_rules(self, obj: Attributed, path: str, identifier_path: Tuple) -> bool:
        return True

    def _check_value_types(self, obj: Attributed, path: str, identifier_path: Tuple, attrib, term: Entity) -> bool:
        valid: bool = True
        value_types = term.get("has_value_type")
        if not value_types:
            value_parsed = isinstance(attrib.value, Attribute)
            if value_parsed or is_curie(attrib.value):
                if value_parsed:
                    value_key = attrib.value.key.split("|")[0]
                else:
                    value_key = attrib.value.split("|")[0]
                if is_curie(value_key):
                    value_term = self.find_term_for(value_key)
                else:
                    value_term = None
                if not value_term or not value_term.is_of_type(term):
                    self.add_warning(
                        obj,
                        path,
                        identifier_path,
                        attrib.key,
                        attrib.value,
                        RequirementLevel.must,
                        f"The value type of {attrib.key} must be a term derived from {attrib.key}, but found {attrib.value}",
                    )
                    valid = False
                    return valid
            else:
                self.add_warning(
                    obj,
                    path,
                    identifier_path,
                    attrib.key,
                    attrib.value,
                    RequirementLevel.must,
                    f"The value type of {attrib.key} must be a term derived from {attrib.key}",
                )
                valid = False
                return valid
        else:
            type_message_errors = []
            for rel in value_types:
                if isinstance(rel.value_type, ListOfType):
                    hit = False
                    for tp in rel.value_type.type_definition.entity.has_value_type:
                        if isinstance(attrib.value, Sequence) and all(
                            isinstance(v, tp.value_type.type_definition) for v in attrib.value
                        ):
                            hit = True
                            break
                    if hit:
                        break
                type_match, message = _is_of_type(attrib, rel)
                if message is not None:
                    type_message_errors.append(message)
                if type_match:
                    break
            else:
                type_message_errors = ", ".join(type_message_errors)
                self.add_warning(
                    obj,
                    path,
                    identifier_path,
                    attrib.key,
                    attrib.value,
                    RequirementLevel.must,
                    f"The value type of {attrib.key} must be a value of type {', '.join([rel.value_type.id for rel in value_types])}, but got {type(attrib.value)} {type_message_errors}",
                )
                valid = False
        return valid

    def _check_units(self, obj: Attributed, path: str, identifier_path: Tuple, attrib, term: Entity) -> bool:
        valid: bool = True
        units = term.get("has_units")
        if units:
            if not isinstance(units, list):
                units = [units]

            is_intensity_measure = _is_intensity_units(units)

            if attrib.group_id is not None:
                try:
                    unit_attrib = obj.get_attribute("UO:0000000|unit", group_identifier=attrib.group_id, raw=True)
                except KeyError:
                    unit_attrib = None
                    if len(units) == 1:
                        logger.warning(f"{attrib.key}'s unit is missing, defaulting to {units[0]}")
                        return valid
            else:
                unit_attrib = None
                if len(units) == 1:
                    logger.debug(f"{attrib.key}'s unit is missing, defaulting to {units[0]}")
                    return valid

            if not unit_attrib and is_intensity_measure:
                try:
                    unit_attrib = obj.get_attribute("MS:1000043|intensity unit", raw=True)
                    if isinstance(unit_attrib, list):
                        unit_attrib = unit_attrib[0]
                except KeyError:
                    pass

            if unit_attrib:
                unit_acc, unit_name = unit_attrib.value.split("|", 1)
                for unit in units:
                    if unit_acc == unit.accession or unit_name == unit.comment:
                        break
                else:
                    self.add_warning(
                        obj,
                        path,
                        identifier_path,
                        attrib.key,
                        attrib.value,
                        RequirementLevel.must,
                        f"The attribute {attrib.key} must have a unit {', '.join([rel.accession + '|' + rel.comment for rel in units])}, but got {unit_acc}|{unit_name}",
                    )
                    valid = False
                    return valid
            else:
                if not term.id in DEFAULT_UNITS:
                    self.add_warning(
                        obj,
                        path,
                        identifier_path,
                        attrib.key,
                        attrib.value,
                        RequirementLevel.must,
                        f"The attribute {attrib.key} must have a unit {', '.join([rel.accession + '|' + rel.comment for rel in units])}, but none were found",
                    )
                    valid = False
                    return valid
        return valid

    def check_attributes(self, obj: Attributed, path: str, identifier_path: Tuple) -> bool:
        valid: bool = True
        for attrib in obj.attributes:
            if attrib.key == "MS:1003276|other attribute value" or attrib.key == "MS:1003275|other attribute name":
                continue
            if self.current_context.visited_attribute(attrib):
                continue
            try:
                acc, name = attrib.key.split("|", 1)
                term = self.find_term_for(acc)
                try:
                    term_by_name = self.find_term_by_name(name)
                except KeyError:
                    term_by_name = None
                if term != term_by_name and term_by_name is not None:
                    self.add_warning(
                        obj,
                        path,
                        identifier_path,
                        attrib.key,
                        attrib.value,
                        RequirementLevel.must,
                        f'{attrib.key}\'s accession {acc} resolved to "{term.id}|{term.name}", '
                        f'but it\'s name "{name}" resolved to "{term_by_name.id}|{term_by_name.name}"',
                    )
                    valid = False
                elif term_by_name is None:
                    self.add_warning(
                        obj,
                        path,
                        identifier_path,
                        attrib.key,
                        attrib.value,
                        RequirementLevel.must,
                        f'{attrib.key}\'s accession {acc} resolved to "{term.id}|{term.name}", '
                        f'but it\'s name "{name}" did not resolve',
                    )
                    valid = False
            except KeyError:
                logger.warn(f"Could not resolve term for {attrib.key} at {path} {identifier_path}")
                valid = False
                self.add_warning(
                    obj,
                    path,
                    identifier_path,
                    attrib,
                    None,
                    RequirementLevel.must,
                    f"Could not resolve term for {attrib.key} at {path} {identifier_path}",
                )
                continue
            except ValueError:
                logger.error(f"Could not parse {attrib.key} at {path} {identifier_path}", exc_info=True)
                valid = False
                self.add_warning(
                    obj,
                    path,
                    identifier_path,
                    attrib,
                    None,
                    RequirementLevel.must,
                    f"Could not parse {attrib.key} at {path} {identifier_path}",
                )

            valid &= self._check_value_types(obj, path, identifier_path, attrib, term)
            valid &= self._check_units(obj, path, identifier_path, attrib, term)

        return valid


@dataclass(frozen=True)
class ValidationError:
    path: str
    identifier_path: Tuple
    attribute: Any
    value: Any
    requirement_level: RequirementLevel
    message: str
    source: str = None

    def __hash__(self):
        return hash((self.path, self.identifier_path, self.attribute, self.message))


class RuleValidator(ValidatorBase):
    name: str
    semantic_rules: List[ScopedSemanticRule]
    object_rules: List[ScopedObjectRuleBase]

    def __init__(self, name, semantic_rules: Optional[List[ScopedSemanticRule]] = None, object_rules: Optional[List[ScopedObjectRuleBase]] = None, error_log: Optional[List] = None):
        super().__init__(error_log)
        self.name = name
        self.semantic_rules = semantic_rules or []
        self.object_rules = object_rules or []

    def apply_rules(self, obj: Attributed, path: str, identifier_path: Tuple) -> bool:
        result = True
        for rule in itertools.chain(self.semantic_rules, self.object_rules):
            if rule.path == path:
                v = rule(obj, path, identifier_path, self)
                level = logging.DEBUG
                if not v and rule.requirement_level > RequirementLevel.may:
                    # level = logging.WARN
                    result &= v
                logger.log(level, f"Applied {rule.id} to {path}:{identifier_path} {v}/{result}")
        return result

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name!r}, {self.semantic_rules}, {self.object_rules})"


class ValidatorChain(ValidatorBase):
    validators: List[ValidatorBase]

    def __init__(self, validators: List[ValidatorBase], *args, **kwargs):
        self.validators = list(validators)
        super().__init__(*args, **kwargs)

    @property
    def error_log(self):
        log = []
        seen = set()
        for validator in self.validators:
            for message in validator.error_log:
                if message not in seen:
                    log.append(message)
                    seen.add(message)
        return log

    @error_log.setter
    def error_log(self, _value):
        pass

    def validate_spectrum(self, spectrum: Spectrum, path: str, library: SpectrumLibrary, parsing_warnings: Optional[List[warnings.WarningMessage]] = None):
        result = True
        for validator in self.validators:
            result &= validator.validate_spectrum(spectrum, path, library, parsing_warnings)
        return result

    def validate_analyte(self, analyte: Analyte, path: str, spectrum: Spectrum, library: SpectrumLibrary):
        result = True
        for validator in self.validators:
            result &= validator.validate_analyte(analyte, path, spectrum, library)
        return result

    def validate_interpretation(self, interpretation: Interpretation, path: str, spectrum: Spectrum, library: SpectrumLibrary):
        result = True
        for validator in self.validators:
            result &= validator.validate_interpretation(interpretation, path, spectrum, library)
        return result

    def apply_rules(self, obj: Attributed, path: str, identifier_path: Tuple) -> bool:
        result = True
        for validator in self.validators:
            result &= validator.apply_rules(obj, path, identifier_path)
        return result

    def check_attributes(self, obj: Attributed, path: str, identifer_path: Tuple) -> bool:
        result = True
        for validator in self.validators:
            result &= validator.check_attributes(obj, path, identifer_path)
        return result

    def reset_context(self):
        for validator in self.validators:
            validator.reset_context()

    def chain(self, validator: ValidatorBase) -> ValidatorBase:
        self.validators.append(validator)
        return self


predicates = {
    "single_spectrum": lambda spec: spec.get_attribute("MS:1003065|spectrum aggregation type") == "MS:1003066|singleton spectrum",
    "consensus_spectrum": lambda spec: spec.get_attribute("MS:1003065|spectrum aggregation type") == "MS:1003066|singleton spectrum",
}


object_rules = {
    "peak_annotations": [SpectrumPeakAnnotationRule()],
    "base": [LibraryFormatVersionFirstRule()]
}


def get_validator_for(name: str) -> RuleValidator:
    """Load a :class:`RuleValidator` with semantic rules for the given rule profile"""
    rules = load_rule_set(name)
    validator = RuleValidator(name, rules)
    return validator


def get_object_validator_for(name: str) -> RuleValidator:
    """Prepare a :class:`RuleValidator` configured with the object rules for the given rule profile"""
    rules = object_rules[name]
    validator = RuleValidator(name, object_rules=rules)
    return validator


def load_default_validator() -> ValidatorChain:
    """Load the core :class:`RuleValidator` and add all the basic object rules for library validation"""
    chain = get_validator_for("base")
    chain |= ControlledVocabularyAttributeValidator()
    chain |= get_object_validator_for("base")
    chain |= get_object_validator_for("peak_annotations")
    return chain



import os
import unittest

from mzspeclib.backends import TextSpectralLibrary
from mzspeclib.cluster import SpectrumCluster

from .common import datafile


class TestSpectrumCluster(unittest.TestCase):

    def get_library(self):
        test_file = datafile("clusters_example.mzlb")
        return TextSpectralLibrary(test_file)

    def test_text_cluster_parsing(self):
        lib = self.get_library()
        cluster: SpectrumCluster = lib.get_cluster(1)

        assert cluster.key == 1
        assert cluster.size == 6


"""Helper module for I/O operations"""

import os
import io
import gzip
import logging

from collections import deque
from urllib import parse as urlparse
from typing import Any, Dict, Iterable, Mapping, Optional, Union

DEFAULT_BUFFER_SIZE = int(2e6)
GZIP_MAGIC = b"\037\213"

GzipFile = gzip.GzipFile
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

try:
    # Fast random acces with Gzip compatibility
    import idzip

    idzip.compressor.IdzipWriter.enforce_extension = False

    GzipFile = idzip.IdzipFile
except ImportError:
    pass


class _LineBuffer(object):
    """
    An implementation detail that treats a stream/iterator over line strings as LIFO
    queue that can have lines pushed back onto it.
    """

    lines: deque
    stream: io.IOBase
    last_line: str
    encoding: Optional[str]
    _stream_is_file_like: bool

    def __init__(
        self,
        stream: io.IOBase,
        lines: Iterable = None,
        last_line: str = None,
        encoding: Optional[str] = None,
    ):
        if lines is None:
            lines = []
        self.lines = deque(lines)
        self.stream = stream
        self.last_line = last_line
        self.encoding = encoding
        self._stream_is_file_like = hasattr(self.stream, "readline")

    def readline(self) -> Union[bytes, str]:
        if self.lines:
            line = self.lines.popleft()
        else:
            line = (
                self.stream.readline()
                if self._stream_is_file_like
                else next(self.stream)
            )
        self.last_line = line
        if self.encoding:
            return line.decode(self.encoding)
        return line

    def push_line(self, line=None):
        if line is None:
            line = self.last_line
            self.last_line = None
        if line is None:
            raise ValueError(
                "Cannot push empty value after the backtrack line is consumed"
            )
        self.lines.appendleft(line)

    def __iter__(self):
        while self.lines:
            line = self.lines.popleft()
            self.last_line = line
            if self.encoding:
                yield line.decode(self.encoding)
            else:
                yield line
        for line in self.stream:
            self.last_line = line
            if self.encoding:
                yield line.decode(self.encoding)
            else:
                yield line

    def __getattr__(self, attr):
        return getattr(self.stream, attr)


def try_cast(value: Any) -> Union[str, int, float, Any]:
    """
    Given a value, if it is a string, attempt to convert it to a numeric type,
    or else return it as is.
    """
    if value is None:
        return value
    if not isinstance(value, str):
        return value
    try:
        return int(value)
    except ValueError:
        pass
    try:
        return float(value)
    except ValueError:
        pass
    return value


def test_gzipped(f) -> bool:
    """
    Checks the first two bytes of the
    passed file for gzip magic numbers

    Parameters
    ----------
    f : file-like or path-like
        The file to test

    Returns
    -------
    bool
    """
    if isinstance(f, os.PathLike):
        f = io.open(f, "rb")
    try:
        current = f.tell()
        assert current >= 0
    except OSError:
        return False
    f.seek(0)
    magic = f.read(2)
    f.seek(current)
    return magic == GZIP_MAGIC


def starts_with_gz_magic(bytestring):
    """
    Test whether or not a byte string starts with
    the GZIP magic bytes.

    Parameters
    ----------
    bytestring : bytes
        The bytes to test.

    Returns
    -------
    bool
    """
    return bytestring.startswith(GZIP_MAGIC)


class _NotClosingWrapper:
    stream: io.BufferedIOBase

    def __init__(self, stream) -> None:
        self.stream = stream

    def __getattr__(self, attrib: str):
        attr = getattr(self.stream, attrib)
        return attr

    def close(self):
        logger.debug("Resetting stream handle %r", self.stream)
        if self.stream.seekable():
            self.stream.seek(0)

    def __iter__(self):
        return iter(self.stream)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        logger.debug("Resetting stream handle %r", self.stream)
        if self.stream.seekable():
            self.stream.seek(0)
        return


def open_stream(
    f: Union[io.IOBase, os.PathLike],
    mode="rt",
    buffer_size: Optional[int] = None,
    encoding: Optional[str] = "utf8",
    newline=None,
    closing=False,
):
    """
    Select the file reading type for the given path or stream.

    Detects whether the file is gzip encoded.
    """
    if isinstance(f, PreBufferedStreamReader):
        if closing:
            return f
        else:
            return _NotClosingWrapper(f)
    if buffer_size is None:
        buffer_size = DEFAULT_BUFFER_SIZE
    is_stream = False
    offset = None
    if "r" in mode:
        if not hasattr(f, "read"):
            f = io.open(f, "rb")
        else:
            is_stream = True
            offset = f.tell()
        if not isinstance(f, io.BufferedReader) and not isinstance(f, io.TextIOWrapper):
            if not closing and is_stream:
                f = _NotClosingWrapper(f)
            buffered_reader = io.BufferedReader(f, buffer_size)
        else:
            if not closing and is_stream:
                f = _NotClosingWrapper(f)
            buffered_reader = f
        if test_gzipped(buffered_reader):
            handle = GzipFile(fileobj=buffered_reader, mode="rb")
        else:
            handle = buffered_reader
    else:
        raise NotImplementedError(
            "Haven't implemented automatic output stream determination"
        )
    try:
        fmode = f.mode
        if isinstance(fmode, int):
            # gzip.GzipFile uses ints in `mode`
            fmode = "b"
    except AttributeError:
        fmode = "b"
    if "b" not in mode and "b" in fmode:
        handle = io.TextIOWrapper(handle, encoding=encoding, newline=newline)
    if is_stream and handle.seekable():
        if offset:
            handle.seek(offset)
    return handle


class CaseInsensitiveDict(Dict[str, Any]):
    """A case sensitive version of a dictionary with string keys."""

    def __init__(self, base=None, **kwargs):
        if base is not None:
            self.update(base)
        if kwargs:
            self.update(kwargs)

    def __getitem__(self, key: str):
        return super().__getitem__(key.lower())

    def __setitem__(self, key: str, value):
        super().__setitem__(key.lower(), value)

    def __delitem__(self, key: str):
        super().__delitem__(key.lower())

    def __contains__(self, __o: str) -> bool:
        if hasattr(__o, "lower"):
            return super().__contains__(__o.lower())
        else:
            return super().__contains__(__o)

    def get(self, key: str, default=None):
        return super().get(key.lower(), default)

    def update(self, value: Mapping[str, Any]):
        super().update({k.lower(): v for k, v in value.items()})


def urlify(path: str) -> str:
    """Convert a path into a URL if it is not already one."""
    parsed = urlparse.urlparse(path)
    if parsed.scheme == "":
        parsed = parsed._replace(scheme="file")
    return urlparse.urlunparse(parsed)


import io
from typing import Literal

DEFAULT_BUFFER_SIZE = 2**18


class PreBufferedStreamReader(io.IOBase):
    """
    A file-like object that can wrap a non-rewindable (e.g. un-seekable)
    file stream like :obj:`sys.stdin` and support seeking over the first **N**
    bytes, but blocking seek operations once the stream has been read beyond
    that point.

    This type exists to adapt :obj:`sys.stdin` or socket files to work with the
    behavior of readers like :class:`~ms_deisotope.data_source.mzml.MzMLLoader`
    that rely on early random access to extract metadata even when full random
    access indexing isn't enabled.

    Attributes
    ----------
    stream : :class:`io.IOBase`
        The object to actually read from. Must support :meth:`io.IOBase.read`,
        :meth:`io.IOBase.tell`, and :meth:`close`. It will also proxy
        :meth:`io.IOBase.fileno` if it is available.
    """

    stream: io.IOBase
    buffer: io.BytesIO

    _cursor: int
    _total_cursor: int
    _buffer_size: int

    def __init__(self, stream, buffer_size=DEFAULT_BUFFER_SIZE):
        self.stream = stream
        dat = self.stream.read(buffer_size)
        self.buffer = io.BytesIO(dat)
        self._cursor = 0
        self._total_cursor = len(dat)
        self._buffer_size = buffer_size

    def read(self, n=-1):
        if n == -1:
            buffer = bytearray()
            buffer.extend(self.buffer.read())
            buffer.extend(self.stream.read())
            self._cursor = self._buffer_size
            self._total_cursor = len(buffer) - self._cursor
            return bytes(buffer)
        buffer = bytearray(n)
        view = memoryview(buffer)
        k = self.buffer.readinto(view)
        self._cursor += k
        j = 0
        if k < n:
            j = self.stream.readinto(view[k:])
            self._total_cursor += j
            k += j
        return bytes(view[:k])

    def fileno(self):
        return self.stream.fileno()

    def seek(self, offset: int, whence: Literal[0, 1, 2] = io.SEEK_SET):
        if whence == io.SEEK_SET and offset > self._buffer_size:
            raise io.UnsupportedOperation("Cannot seek beyond the pre-loaded buffer")
        elif whence == io.SEEK_CUR or whence == io.SEEK_END:
            raise io.UnsupportedOperation("Cannot seek not from the start of the file")
        if self._total_cursor > self._buffer_size:
            raise io.UnsupportedOperation("Already read beyond the pre-loaded buffer")
        else:
            self.buffer.seek(offset, whence)
            self._cursor = offset

    def truncate(self, size=None):
        raise io.UnsupportedOperation("Read-only")

    def close(self):
        self.buffer.close()
        return self.stream.close()

    @property
    def closed(self):
        return self.stream.closed

    def seekable(self):
        return True

    def tell(self):
        if self._cursor < self._buffer_size:
            return self._cursor
        return self._total_cursor

    def writable(self):
        return False

    def __enter__(self):
        self.stream.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



"""Abstract base types for spectral library indices"""
import io
import pathlib
import warnings

from typing import Collection, Iterator, Optional, Tuple, Union, Any, List, NamedTuple


class IndexInitializedRecord(NamedTuple):
    """
    An instantiated index object, and whether it was initialized already or the caller
    needs to populate it.

    Attributes
    ----------
    index : :class:`~.IndexBase`
        A newly index object
    initialized : :class:`bool`
        Whether the index was populated or not
    """

    index: 'IndexBase'
    initialized: bool


class IndexRecordBase:
    """
    A base type describing index entries

    Attributes
    ----------
    number : int
        The "key" of a library entry
    index : int
        The sequential index for the entry
    offset : int
        The offset into the file for the entry. Usually in bytes but
        for some file types which use a container format that abstracts
        the IO, this might just be the index again.
    """

    __slots__ = ()

    number: int
    index: int
    offset: int
    name: str


class IndexBase(Collection):
    """
    A base type for spectral indices.

    Retrieve information about entries' identifiers and any associated
    metadata.
    """

    @classmethod
    def from_filename(cls, filename: Union[str, pathlib.Path, io.FileIO], library=None) -> IndexInitializedRecord:
        """
        Get a file path for an index file, given the library filename.

        Returns
        -------
        str or None
        """
        raise NotImplementedError()

    @classmethod
    def exists(cls, filename: Union[str, pathlib.Path, io.FileIO]) -> bool:
        """Check if an index file exists"""
        return False

    def offset_for(self, record_label) -> int:
        """Retrieve the byte offset of a spectrum identifier"""
        record = self.record_for(record_label)
        return record.offset

    def offset_for_cluster(self, record_label) -> int:
        """Retrieve the byte offset of a cluster identifier"""
        record = self.record_for_cluster(record_label)
        return record.offset

    def record_for(self, record_label: Union[int, str]) -> IndexRecordBase:
        """Retrieve a an index record for a spectrum identifier"""
        record = self.search(record_label)
        if isinstance(record, list):
            warnings.warn(
                f"Multiple records found for {record_label}, using the first")
            record = record[0]
        return record

    def record_for_cluster(self, record_label: int) -> IndexRecordBase:
        """Retrieve a an index record for a cluster identifier"""
        record = self.search_clusters(record_label)
        if isinstance(record, list):
            warnings.warn(
                f"Multiple records found for {record_label}, using the first")
            record = record[0]
        return record

    def search(self, i: Union[str, int, slice], **kwargs) -> Union[IndexRecordBase, List[IndexRecordBase]]:
        """Search for one or more spectrum records by index, slice or identifier"""
        raise NotImplementedError()

    def search_clusters(self, i: Optional[Union[int, slice]]=None, **kwargs) -> Union[IndexRecordBase, List[IndexRecordBase]]:
        """Search for one or more cluster records by index, slice or identifier"""
        raise NotImplementedError()

    def add(self, number: int, offset: int, name: str, analyte: Any, attributes=None):
        """
        Add a new entry to the spectrum index.

        Parameters
        ----------
        number : int
            A numerical identifier for this spectrum.
        offset : int
            The offset in the file to reach the spectrum (in bytes if appropriate)
        name : str,
            A text identifier for this spectrum.
        analyte : str, optional
            A text representation of the analyte for that record
        attributes : Dict[str, Any], optional
            A key-value pair collection of this record, currently not supported.
        """
        raise NotImplementedError()

    def add_cluster(self, number: int, offset: int, attributes=None):
        """
        Add a new entry to the cluster index.

        Parameters
        ----------
        number : int
            A numerical identifier for this spectrum.
        offset : int
            The offset in the file to reach the spectrum (in bytes if appropriate)
        attributes : Dict[str, Any], optional
            A key-value pair collection of this record, currently not supported.
        """
        raise NotImplementedError()

    def commit(self):
        """
        Commit any index state to disk, if this index supports persistence.

        Has no effect on index types that do not have a persistence functionality.
        """
        raise NotImplementedError()

    def iter_clusters(self) -> Iterator[IndexRecordBase]:
        """Iterate over cluster records"""
        raise NotImplementedError()

    def iter_spectra(self) -> Iterator[IndexRecordBase]:
        """Iterate over peptide records"""
        for i in range(len(self)):
            yield self[i]

    def _get_by_index(self, i: Union[int, slice]) -> Union[IndexRecordBase, List[IndexRecordBase]]:
        raise NotImplementedError()

    def __iter__(self):
        return self.iter_spectra()

    def __getitem__(self, i: Union[int, str, slice]):
        return self.search(i)

    def __len__(self):
        raise NotImplementedError()

    def __contains__(self, key) -> bool:
        try:
            hit = self.search(key)
            return True
        except (KeyError, IndexError, ValueError):
            return False

    def check_names_unique(self) -> bool:
        """
        Check that all indexed spectra have unique ``spectrum name`` parameters.

        Returns
        -------
        bool:
            Whether the spectrum names in the index are unique.
        """
        seen = set()
        for record in self:
            if record.name in seen:
                return False
            seen.add(record.name)
        return True



from __future__ import print_function

import textwrap

from typing import Any, Dict,  List, Optional, TYPE_CHECKING

from mzspeclib.attributes import (
    AttributeManager, AttributeManagedProperty, AttributeListManagedProperty,
    AttributeProxy as _AttributeProxy, AttributeFacet
)
from mzspeclib.analyte import Analyte, InterpretationCollection, Interpretation
from .const import (SPECTRUM_NAME, LIBRARY_SPECTRUM_INDEX, LIBRARY_SPECTRUM_KEY, PRECURSOR_MZ, CHARGE_STATE,
                    SPECTRUM_AGGREGATION_TYPE, NUMBER_OF_REPLICATE_SPECTRA_USED, NUMBER_OF_REPLICATE_SPECTRA_AVAILABLE,
                    THEORETICAL_MZ, PEAK_ATTRIBUTE)

if TYPE_CHECKING:
    from mzspeclib.spectrum_library import SpectrumLibrary

#A class that holds data for each spectrum that is read from the SpectralLibrary class


class SpectrumAggregation(_AttributeProxy):
    aggregation_type = AttributeManagedProperty(SPECTRUM_AGGREGATION_TYPE)
    replicates_used = AttributeManagedProperty[int](NUMBER_OF_REPLICATE_SPECTRA_USED)
    replicates_available = AttributeManagedProperty[int](NUMBER_OF_REPLICATE_SPECTRA_AVAILABLE)


class Spectrum(AttributeManager):
    """
    A mass spectrum stored in a spectral library.

    The :class:`Spectrum` type is a :class:`~.AttributeManager`, interacting
    with the spectrum's attribute list itself. It also encloses :class:`~.Interpretation`
    and :class:`~.Analyte` members which have their own attributes.

    Attributes
    ----------
    peak_list : List[Tuple[float, float, Optional[List[:class:`mzpaf.PeakAnnotation`]], Optional[List[Any]]]]
        The list of peaks which may or may not be annotated and may or may not have
        aggregation metrics.
    interpretations : :class:`~.InterpretationCollection`
        The set of all interpretations associated with this spectrum, and their analytes.
    """

    peak_list: List
    analytes: Dict[str, Analyte]
    interpretations: InterpretationCollection
    _source: Optional['SpectrumLibrary']

    #### Constructor
    def __init__(self, attributes=None, peak_list=None, analytes=None,
                 interpretations=None):
        """

        Parameters
        ----------
        attributes : list
            A list of attribute [key, value (, group)] sets to initialize to.
        peak_list : list
            A list of tuples representing (annotated) peaks
        analytes : dict[str, :class:`~.Analyte`]
            A mapping from identifier to :class:`~.Analyte` unique within this
            :class:`Spectrum`.
        interpretations : :class:`~.InterpretationCollection`
            A mapping from identifier to :class:`~.Interpretation` unique within
            this :class:`Spectrum`.
        """
        if peak_list is None:
            peak_list = []
        if analytes is None:
            analytes = dict()
        if interpretations is None:
            interpretations = InterpretationCollection()
        else:
            interpretations = InterpretationCollection(interpretations)
        super(Spectrum, self).__init__(attributes)
        self.peak_list = peak_list
        self.analytes = analytes
        self.interpretations = interpretations

    name = AttributeManagedProperty[str](SPECTRUM_NAME)
    key = AttributeManagedProperty[int](LIBRARY_SPECTRUM_KEY)
    index = AttributeManagedProperty[int](LIBRARY_SPECTRUM_INDEX)

    precursor_mz = AttributeListManagedProperty[float](
        [PRECURSOR_MZ,
         THEORETICAL_MZ])

    _precursor_charge = AttributeManagedProperty[int](CHARGE_STATE)

    spectrum_aggregation = AttributeFacet[SpectrumAggregation](SpectrumAggregation)
    peak_aggregations = AttributeManagedProperty(PEAK_ATTRIBUTE, multiple=True)

    @property
    def precursor_charge(self) -> int:
        """Obtain the spectrum's precursor ion charge or analyte charge"""
        if self._precursor_charge:
            return self._precursor_charge
        for analyte in self.analytes.values():
            if analyte.charge:
                return analyte.charge

    def add_analyte(self, analyte: Analyte):
        """Add an :class:`~.Analyte` to the spectrum"""
        self.analytes[str(analyte.id)] = analyte

    def get_analyte(self, analyte_id: str) -> Analyte:
        """Get an :class:`~.Analyte` by ID from the spectrum"""
        return self.analytes[str(analyte_id)]

    def remove_analyte(self, analyte_id: str):
        """Remove an :class:`~.Analyte` by ID from the spectrum"""
        analyte_id = str(analyte_id)
        del self.analytes[analyte_id]
        interpretation: Interpretation
        for interpretation in self.interpretations.values():
            if interpretation.has_analyte(analyte_id):
                interpretation.remove_analyte(analyte_id)

    def add_interpretation(self, interpretation: Interpretation):
        self.interpretations.add_interpretation(interpretation)

    def get_interpretation(self, interpretation_id: str) -> Interpretation:
        return self.interpretations.get_interpretation(interpretation_id)

    def __eq__(self, other):
        result = super(Spectrum, self).__eq__(other)
        if result:
            result = self.peak_list == other.peak_list
        if result:
            result = self.analytes == other.analytes
        return result

    def __repr__(self):  # pragma: no cover
        template = f"{self.__class__.__name__}("
        lines = list(map(str, self.attributes))
        if not lines:
            template += "[], "
        else:
            template += "[\n%s], " % textwrap.indent(',\n'.join(lines), ' ' * 2)
        lines = list(map(str, self.peak_list))
        if not lines:
            template += "peak_list=[])"
        else:
            template += "peak_list=[\n%s])" % textwrap.indent(
                ',\n'.join(lines), ' ' * 2)
        return template

    def __str__(self):  # pragma: no cover
        return self.write("text")

    def write(self, format="text", **kwargs):  # pragma: no cover
        """
        Write out the spectrum in any of the supported formats

        Parameters
        ----------
        format : str
            The name of the format to write in
        **kwargs
            Passed to implementation
        """
        #### Set a buffer to fill with string data
        buffer = ''

        #### Make the format string lower case to facilitate comparisons
        format = format.lower()

        #### If the format is text
        if format == "text":
            from mzspeclib.backends.text import format_spectrum
            return format_spectrum(self, **kwargs)

        #### If the format is TSV
        elif format == "tsv" or format == "csv":

            #### Set the appropriate delimiter for format
            delimiter = "\t"
            if format == "csv":
                delimiter = ","

            #### Create the header line and columns line
            buffer += "# Spectrum attributes\n"
            buffer += delimiter.join("cv_param_group", "accession", "name", "value_accession", "value\n")

            for attribute in self.attributes:
                if len(attribute) == 2:
                    key,value = attribute
                    cv_param_group = ''
                elif len(attribute) == 3:
                    key,value,cv_param_group = attribute
                    if format == "csv" and ',' in str(value):
                        value = '"' + value + '"'
                else:
                    print("ERROR: Unsupported number of items in attribute")
                    print(attribute)
                    raise ValueError(
                        f"Unsupported number of items in attribute: {attribute}")
                components = key.split('|',1)
                if len(components) == 2:
                    accession,name = components
                else:
                    print("ERROR: Unsupported number of items in components")
                    print(components)
                    raise ValueError(f"Unsupported number of items in components: {components}")
                components = str(value).split('|',1)
                if len(components) == 2:
                    value_accession,value = components
                    value = str(value)
                    if format == "csv" and ',' in value:
                        value = '"' + value + '"'
                elif len(components) == 1:
                    value = str(value)
                    if format == "csv" and ',' in value:
                        value = '"' + value + '"'
                    value_accession = ''
                else:
                    print("ERROR: Unsupported number of items in components")
                    print(components)
                    raise ValueError(
                        f"Unsupported number of items in components: {components}")

                #### Create the data line
                buffer += delimiter.join(map(str, [cv_param_group,accession,name,value_accession,value]))+"\n"

            #### Create the header line and columns line
            buffer += "# Peak list\n"
            buffer += "mz\tintensity\tinterpretation\n"

            #### Write out the peak list
            for peak in self.peak_list:
                mz,intensity,interpretation = peak
                if format == "csv" and ',' in str(interpretation):
                    interpretation = '"' + interpretation + '"'
                buffer += delimiter.join(map(str, [mz,intensity,interpretation]))+"\n"

        #### If the format is JSON
        elif format == "json":
            from mzspeclib.backends.json import format_spectrum
            return format_spectrum(self, **kwargs)

        #### Otherwise we don't know this format
        else:
            raise ValueError(f"ERROR: Unrecogized format '{format}'")

        return(buffer)



import enum

from pprint import pformat
from numbers import Number
from typing import List, Union, Mapping, Sequence, Iterator

import numpy as np

PeakDtype = np.dtype([('mz', np.float64), ('intensity', np.float32),
                      ('annotation', object), ('aggregation', object)])

PeakType = Mapping[str, Union[float, list]]

class ErrorUnit(enum.Enum):
    Da = 'da'
    PPM = 'ppm'


class PeakList(Sequence[PeakType]):
    peaks: np.ndarray

    def __init__(self, peaks):
        if isinstance(peaks, np.ndarray) and peaks.dtype == PeakDtype:
            self.peaks = peaks
        else:
            self.peaks = np.array([tuple(p) for p in peaks], dtype=PeakDtype)

    def __len__(self):
        return len(self.peaks)

    def __getitem__(self, i):
        return self.peaks[i]

    def __iter__(self) -> Iterator[PeakType]:
        return iter(self.peaks)

    def __repr__(self):
        return f"{self.__class__.__name__}({pformat(self.peaks, indent=2)})"

    def __eq__(self, other):
        if other is None:
            return False
        if not isinstance(other, PeakList):
            other = PeakList(other)
        valid = np.allclose(self.peaks['mz'], other.peaks['mz'])
        if not valid:
            return False
        valid = np.allclose(self.peaks['intensity'], other.peaks['intensity'])
        if not valid:
            return False
        return True

    def __ne__(self, other):
        return not self == other

    def find(self, mz, error_tolerance=10, error_unit='ppm') -> List[Sequence[PeakType]]:
        if isinstance(mz, Number):
            mz = [mz]
        mzs = self.peaks['mz']
        ii = np.searchsorted(mzs, mz)
        n = len(self)
        error_unit = ErrorUnit(error_unit)

        outs = []
        if error_unit == ErrorUnit.Da:
            for i, mz_i in zip(ii, mz):
                low = i - 1
                while low >= 0:
                    if abs(mzs[low] - mz_i) < error_tolerance:
                        low -= 1
                    else:
                        low += 1
                        break
                high = i
                while high < n:
                    if abs(mzs[high] - mz_i) < error_tolerance:
                        high += 1
                    else:
                        high -= 1
                        break
                outs.extend(self.peaks[slice(low, high + 1)])
        elif error_unit == ErrorUnit.PPM:
            error_tolerance /= 1e6
            for i, mz_i in zip(ii, mz):
                low = i - 1
                while low >= 0:
                    if abs(mzs[low] - mz_i) / mz_i < error_tolerance:
                        low -= 1
                    else:
                        low += 1
                        break
                high = i
                while high < n:
                    if abs(mzs[high] - mz_i) / mz_i < error_tolerance:
                        high += 1
                    else:
                        high -= 1
                        break
                outs.extend(self.peaks[slice(low, high + 1)])
        return outs



from __future__ import print_function

from dataclasses import dataclass

from typing import Dict,  List

from mzspeclib.attributes import AttributeManager, AttributeManagedProperty, AttributeGroupFacet
from .utils import ensure_iter, flatten

SIMILAR_SPECTRUM_KEYS = "MS:1003263|similar spectrum keys"
SIMILAR_SPECTRUM_USI = "MS:1003264|similar spectrum USI"

CLUSTER_KEY = "MS:1003267|spectrum cluster key"
CLUSTER_SIZE = "MS:1003320|spectrum cluster size"

CLUSTER_MEMBERS_KEYS = "MS:1003268|spectrum cluster member spectrum keys"
CLUSTER_MEMBER_USI = "MS:1003269|spectrum cluster member USI"

CLUSTER_SUMMARY_STATS = "MS:1003321|summary statistics of clustered spectra"


@dataclass
class ClusterMemberRef:
    pass


@dataclass
class SpectrumRef(ClusterMemberRef):
    key: str


@dataclass
class USIRef(ClusterMemberRef):
    usi: str

    @property
    def key(self) -> str:
        return self.usi


class SpectrumCluster(AttributeManager):
    def __init__(self, attributes: List=None):
        super().__init__(attributes)

    key = AttributeManagedProperty[int](CLUSTER_KEY)
    _member_references = AttributeManagedProperty(CLUSTER_MEMBERS_KEYS)
    _cluster_member_usis = AttributeManagedProperty(CLUSTER_MEMBER_USI)

    size = AttributeManagedProperty[int](CLUSTER_SIZE)

    @property
    def members(self) -> List[ClusterMemberRef]:
        internal_refs = [
            SpectrumRef(k) for k in flatten(ensure_iter(self._member_references))
        ]
        usi_members = [USIRef(k) for k in ensure_iter(self._cluster_member_usis)]
        return internal_refs + usi_members



import unittest

from mzspeclib.spectrum_library import SpectrumLibrary
from mzspeclib.validate import validator

from .common import datafile


class TestLibrarySemanticValidator(unittest.TestCase):
    def open_library(self):
        test_file = datafile("chinese_hamster_hcd_selected_head.mzlb.txt")
        library = SpectrumLibrary(filename=test_file)
        return library

    def validate_library(self, valid):
        library = self.open_library()
        return valid.validate_library(library)

    def test_validate_base(self):
        valid = validator.get_validator_for("base")
        assert self.validate_library(valid)

    def test_validate_peptide(self):
        valid = validator.get_validator_for("base")
        valid = valid.chain(validator.get_validator_for("peptide"))
        assert self.validate_library(valid)

    def test_validate_silver(self):
        valid = validator.get_validator_for("base")
        valid = valid.chain(validator.get_validator_for("silver"))
        assert not self.validate_library(valid)

    def test_validate_peak_annotations(self):
        valid = validator.get_validator_for("base") | validator.get_object_validator_for("peak_annotations")
        assert self.validate_library(valid)

        test_file = datafile("bad_peak_annotations.mzlb.txt")
        library = SpectrumLibrary(filename=test_file)
        assert not valid.validate_library(library)



"""
Read DIA-NN TSV spectral libraries.

For more information, see `DIA-NN's website <https://github.com/vdemichev/DiaNN`_
"""
import json
import os

from typing import List, Optional, Tuple, Dict, Iterator, Any, Union

from pyteomics import proforma

from mzspeclib import annotation
from mzspeclib.backends.base import DEFAULT_VERSION, FORMAT_VERSION, LIBRARY_NAME, _CSVSpectralLibraryBackendBase
from mzspeclib.backends.utils import open_stream, urlify
from mzspeclib.spectrum import Spectrum, SPECTRUM_NAME
from mzspeclib.const import (
    PROFORMA_SEQ as PROFORMA_PEPTIDE_TERM,
    PROFORMA_ION,
    STRIPPED_PEPTIDE_SEQ as STRIPPED_PEPTIDE_TERM,
    SELECTED_ION_MZ as SPECTRUM_SELECTED_ION_MZ,
    CHARGE_STATE,
    SOURCE_FILE,
    CUSTOM_ATTRIBUTE_NAME,
    CUSTOM_ATTRIBUTE_VALUE
)


def _rewrite_unimod_peptide_as_proforma(sequence: str) -> str:
    return sequence.replace("(", '[').replace(')', ']').replace("UniMod", "UNIMOD")


THEO_SELECTED_ION_MZ = "MS:1003053|theoretical monoisotopic m/z"

NO_LOSS = 'noloss'


def _parse_value(value: str) -> Union[float, int, str, bool]:
    try:
        return json.loads(value)
    except json.JSONDecodeError:
        lower = value.lower()
        if lower == "true":
            return True
        elif lower == "false":
            return False
        return value


class DIANNTSVSpectralLibrary(_CSVSpectralLibraryBackendBase):
    """Reader for DIA-NN TSV spectral libraries."""

    format_name = "dia-nn.tsv"

    _custom_spectrum_keys = [
        "ExcludeFromAssay",
        "AllowForNormalization",
    ]

    _custom_analyte_keys = [
        "Proteotypic",
        "ProteinGroup",
    ]

    _required_columns = ['transition_group_id', 'PrecursorMz', 'PrecursorCharge',
                         'FullUniModPeptideName', 'ProductMz', 'LibraryIntensity',
                         'FragmentType', 'FragmentSeriesNumber',
                         'FragmentCharge', 'FragmentLossType']

    def __init__(self, filename: str, index_type=None, **kwargs):
        super().__init__(filename, index_type=index_type, delimiter='\t', **kwargs)

    def _spectrum_origin_type(self):
        key = "MS:1003072|spectrum origin type"
        value = "MS:1003074|predicted spectrum"
        return key, value

    def _spectrum_aggregation_type(self):
        key = "MS:1003065|spectrum aggregation type"
        value = "MS:1003074|predicted spectrum"
        return key, value

    def read_header(self) -> bool:
        result = super().read_header()
        self.add_attribute(FORMAT_VERSION, DEFAULT_VERSION)
        if hasattr(self.filename, 'name'):
            name = self.filename.name.replace(".gz", '').rsplit('.', 1)[0].split(os.sep)[-1]
        else:
            name = self.filename.replace(".gz", '').rsplit(".", 1)[0].split(os.sep)[-1]
        self.add_attribute(LIBRARY_NAME, name)
        self.add_attribute("MS:1003207|library creation software", "MS:1003253|DIA-NN")
        return result

    def create_index(self):
        with open_stream(self.filename, 'rb') as stream:
            header = stream.readline()
            header_cols = header.split(b'\t')
            column_key = header_cols.index(b'transition_group_id')
            offset = stream.tell()

            line = stream.readline()
            tokens = line.split(b'\t')
            key = tokens[column_key]
            index = {key: offset}
            n = 0
            self.index.add(
                number=n,
                offset=offset,
                name=key.decode("utf8"),
                analyte=None
            )
            n += 1

            # To hold previous values
            last_offset = None
            last_key = None
            while line:
                tokens = line.split(b'\t')
                if tokens[column_key] != key:
                    key = tokens[column_key]
                    self.index.add(
                        number=n,
                        offset=offset,
                        name=key.decode("utf8") ,
                        analyte=None
                    )
                    n += 1
                    last_offset = stream.tell()
                else:
                    offset = stream.tell()
                line = stream.readline()
        # self.index.add(
        #     number=n,
        #     offset=last_offset,
        #     name=tokens[column_key].decode('utf8'),
        #     analyte=None
        # )
        n += 1
        self.index.commit()
        return n

    def _parse_from_buffer(self, buffer: List[Dict[str, Any]], spectrum_index: Optional[int] = None) -> Spectrum:
        spec = self._new_spectrum()
        descr = buffer[0]

        spec.add_attribute(SPECTRUM_NAME, descr['transition_group_id'])
        spec.add_attribute(SPECTRUM_SELECTED_ION_MZ, float(descr['PrecursorMz']))

        if 'FileName' in descr:
            spec.add_attribute(SOURCE_FILE, urlify(descr['FileName']))

        if 'decoy' in descr and int(descr['decoy']):
            spec.add_attribute("MS:1003072|spectrum origin type", "MS:1003192|decoy spectrum")
        else:
            spec.add_attribute(*self._spectrum_origin_type())

        spec.add_attribute(*self._spectrum_aggregation_type())

        if 'IonMobility' in descr:
            spec.add_attribute("MS:1002476|ion mobility drift time", float(descr['IonMobility']))

        analyte = self._new_analyte('1')

        pf_seq = _rewrite_unimod_peptide_as_proforma(descr['FullUniModPeptideName'])
        peptide = proforma.ProForma.parse(pf_seq)

        if 'PeptideSequence' in descr:
            analyte.add_attribute(STRIPPED_PEPTIDE_TERM, descr['PeptideSequence'])
        peptide.charge_state = descr['PrecursorCharge']
        analyte.add_attribute(PROFORMA_ION, str(peptide))
        analyte.add_attribute("MS:1001117|theoretical mass", peptide.mass)
        spec.add_attribute(CHARGE_STATE, int(descr['PrecursorCharge']))

        protein_group_id = analyte.get_next_group_identifier()
        if "UniprotID" in descr:
            analyte.add_attribute(
                "MS:1000885|protein accession",
                descr['UniprotID'],
                group_identifier=protein_group_id
            )
        if "ProteinName" in  descr:
            if descr['ProteinName']:
                analyte.add_attribute(
                    "MS:1000886|protein name",
                    descr["ProteinName"],
                    group_identifier=protein_group_id
                )

        for key in self._custom_analyte_keys:
            if key in descr:
                analyte.add_attribute_group([
                    [CUSTOM_ATTRIBUTE_NAME, key],
                    [CUSTOM_ATTRIBUTE_VALUE, _parse_value(descr[key])]
                ])

        spec.add_analyte(analyte)
        spec.peak_list = self._generate_peaks(buffer)
        spec.add_attribute("MS:1003059|number of peaks", len(spec.peak_list))

        for key in self._custom_spectrum_keys:
            if key in descr:
                spec.add_attribute_group([
                    [CUSTOM_ATTRIBUTE_NAME, key],
                    [CUSTOM_ATTRIBUTE_VALUE, _parse_value(descr[key])]
                ])

        if spectrum_index is not None:
            spec.index = spectrum_index
        else:
            spec.index = -1
        spec.key = spec.index + 1
        return spec

    def _generate_peaks(self, batch: List[Dict[str, Any]]) -> List[Tuple[float, float, List[annotation.IonAnnotationBase], List]]:
        peaks = []
        for row in batch:
            mz = float(row['ProductMz'])
            intensity = float(row['LibraryIntensity'])

            series = row['FragmentType']
            ordinal = int(row['FragmentSeriesNumber'])
            charge = int(row['FragmentCharge'])

            loss_type = row['FragmentLossType']
            if loss_type != NO_LOSS:
                loss_type = annotation.NeutralName.parse('-' + loss_type)
            else:
                loss_type = None

            annot = annotation.PeptideFragmentIonAnnotation(
                series, ordinal, neutral_losses=loss_type, charge=charge,
                mass_error=annotation.MassError(0, 'Da')
            )

            peak = [
                mz, intensity, [annot], []
            ]
            peaks.append(peak)
        return peaks

    def _batch_rows(self, iterator: Iterator[Dict[str, Any]]) -> Iterator[List[Dict[str, Any]]]:
        group_key = None
        group = []
        for row in iterator:
            key = row['transition_group_id']
            if group_key is None:
                group_key = key
                group.append(row)
            elif group_key == key:
                group.append(row)
            else:
                yield group
                group = [row]
                group_key = key
        if group:
            yield group


DiaNNTSVSpectralLibrary = DIANNTSVSpectralLibrary




import logging

from typing import TYPE_CHECKING, List, Tuple

from mzspeclib.attributes import Attributed

from mzspeclib.spectrum import Spectrum
from mzspeclib.annotation import IonAnnotationBase, InvalidAnnotation

from mzspeclib.validate.level import RequirementLevel
from mzspeclib.utils import ValidationWarning

if TYPE_CHECKING:
    from .validator import ValidatorBase
    from mzspeclib.spectrum_library import SpectrumLibrary

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class ScopedObjectRuleBase:
    """
    A validation rule that cannot be expressed in terms of a semantic attribute constraint.

    Attributes
    ----------
    id : str
        A unique identifier for this rule
    path : str
        The validation path this rule applies to, e.g. /Library or /Library/Spectrum/Analyte
    requirement_level : :class:`~.RequirementLevel`
        How strong the requirement this rule be obeyed is
    """

    id: str
    path: str
    requirement_level: RequirementLevel

    def __init__(self, id, path: str, requirement_level: RequirementLevel=RequirementLevel.should) -> None:
        self.id = id
        self.path = path
        self.requirement_level = requirement_level

    def __eq__(self, other):
        if other is None:
            return False
        props_eq = self.id == other.id and self.path == other.path and self.requirement_level == other.requirement_level
        if not props_eq:
            return False
        return isinstance(other, self.__class__) or isinstance(self, other.__class__)

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.id)

    def __call__(self, obj: Attributed, path: str, identifier_path: Tuple, validator_context: "ValidatorBase") -> bool:
        return self.validate(obj, path, identifier_path, validator_context)

    def validate(self, obj: Attributed, path: str, identifier_path: Tuple, validator_context: "ValidatorBase") -> bool:
        raise NotImplementedError()


class LibraryFormatVersionFirstRule(ScopedObjectRuleBase):
    def __init__(self, requirement_level: RequirementLevel = RequirementLevel.must):
        super().__init__("Library_format_version_first_rule", "/Library", requirement_level=requirement_level)

    def validate(self, obj: "SpectrumLibrary", path: str, identifier_path: Tuple, validator_context: "ValidatorBase") -> bool:
        errors = []
        attr = next(iter(obj.attributes))
        is_format_version_key = attr.key == "MS:1003186|library format version"
        if not is_format_version_key:
            validator_context.add_warning(
                obj,
                path,
                identifier_path,
                self,
                attr,
                self.requirement_level,
                f"The first attribute of the library is not 'MS:1003186|library format version': {attr}",
            )
        if errors:
            return False
        return True


class SpectrumPeakAnnotationRule(ScopedObjectRuleBase):
    def __init__(self, requirement_level: RequirementLevel=RequirementLevel.should):
        super().__init__(
            "Spectrum_peak_annotation_rule", "/Library/Spectrum", requirement_level=requirement_level)

    def validate(self, obj: Spectrum, path: str, identifier_path: Tuple, validator_context: "ValidatorBase") -> bool:
        errors = []
        for i, peak in enumerate(obj.peak_list):
            annotations = peak[2]
            annotations: List[IonAnnotationBase]
            for annot in annotations:
                if isinstance(annot, InvalidAnnotation):
                    errors.append(annot)
                    validator_context.add_warning(
                        obj,
                        path,
                        identifier_path,
                        self,
                        annot,
                        self.requirement_level,
                        f"Invalid peak annotation for peak {i} of Spectrum={obj.key} {annot.content}"
                    )
                else:
                    analyte_id = annot.analyte_reference
                    if analyte_id is not None:
                        if analyte_id not in obj.analytes:
                            validator_context.add_warning(
                                obj,
                                path,
                                identifier_path,
                                self,
                                annot,
                                self.requirement_level,
                                (f"Analyte of peak annotation {annot} for peak {i} of "
                                 f"Spectrum={obj.key} {annot.analyte_reference} is missing")
                            )


        if errors:
            return False
        return True
